#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/protocol/iccp.py
# DESCRIPTION:    Internal Component Control Protocol
# CREATED:        16.11.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin Internal Component Control Protocol


"""

from __future__ import annotations
from typing import Dict, List, Union
import uuid
import traceback
from struct import pack, unpack
from enum import IntEnum, Enum
from firebird.base.types import ANY
from firebird.base.signal import eventsocket
from firebird.base import protobuf
from firebird.base.config import ZMQAddress, Config, ConfigProto, PROTO_CONFIG
from saturnin.base import InvalidMessageError, StopError, PROTO_PEER, PeerDescriptor, \
     Channel, Protocol, Message, Session, RoutingID, TZMQMessage, Outcome

class MsgType(IntEnum):
    """Control message type.
    """
    READY = 1
    REQUEST = 2
    OK = 3
    ERROR = 4
    STOP = 5
    FINISHED = 6

class Request(Enum):
    """Service Controller Request Codes.
    """
    CONFIGURE = b'CONF'

class ICCPMessage(Message):
    """Service Control Message.
    """
    def __init__(self):
        #: Type of message
        self.msg_type: MsgType = None
    def __str__(self):
        return f"{self.__class__.__qualname__}[{'NONE' if self.msg_type is None else self.msg_type.name}]"
    __repr__ = __str__
    def from_zmsg(self, zmsg: TZMQMessage) -> None:
        """Populate message data from sequence of ZMQ data frames.

        Arguments:
            zmsg: Sequence of frames that should be deserialized.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.
        """
        try:
            self.msg_type = MsgType(unpack('!H', zmsg[0])[0])
            if self.msg_type is MsgType.READY:
                proto = protobuf.create_message(PROTO_PEER, zmsg[1])
                self.peer = PeerDescriptor.from_proto(proto)
                proto = protobuf.create_message(protobuf.PROTO_STRUCT, zmsg[2])
                self.endpoints = {}
                for k, v in protobuf.struct2dict(proto).items():
                    for i in range(len(v)):
                        v[i] = ZMQAddress(v[i])
                    self.endpoints[k] = v
            elif self.msg_type is MsgType.ERROR:
                self.error = zmsg[1].decode('utf8')
            elif self.msg_type is MsgType.FINISHED:
                self.outcome = Outcome(zmsg[1].decode())
                self.details = [v.decode('utf8', errors='replace') for v in zmsg[2:]]
            elif self.msg_type is MsgType.REQUEST:
                self.request = Request(zmsg[1])
                if self.request is Request.CONFIGURE:
                    self.config = protobuf.create_message(PROTO_CONFIG, zmsg[2])
        except Exception as exc:
            raise InvalidMessageError("Invalid message") from exc
    def as_zmsg(self) -> TZMQMessage:
        """Returns message as sequence of ZMQ data frames.
        """
        try:
            zmsg = [pack('!H', self.msg_type)]
            if self.msg_type is MsgType.READY:
                zmsg.append(self.peer.as_proto().SerializeToString())
                zmsg.append(protobuf.dict2struct(self.endpoints).SerializeToString())
            elif self.msg_type is MsgType.ERROR:
                zmsg.append(self.error.encode('utf8', errors='replace'))
            elif self.msg_type is MsgType.FINISHED:
                zmsg.append(self.outcome.value.encode('utf-8', errors='replace'))
                zmsg.extend(v.encode('utf-8', errors='replace') for v in self.details)
            elif self.msg_type is MsgType.REQUEST:
                if self.request is Request.CONFIGURE:
                    zmsg.append(self.config.SerializeToString())
        except Exception:
            traceback.print_exc()
        return zmsg
    def clear(self) -> None:
        """Clears message data.
        """
        self.msg_type = None
        for attr in ('peer', 'endpoints', 'error', 'request', 'config'):
            if hasattr(self, attr):
                delattr(self, attr)
    def copy(self) -> Message:
        """Returns copy of the message.
        """
        msg = self.__class__()
        msg.msg_type = self.msg_type
        if self.msg_type is MsgType.READY:
            msg.peer = self.peer.copy()
            msg.endpoints = self.endpoints.copy()
        elif self.msg_type is MsgType.ERROR:
            msg.error = self.error
        elif self.msg_type is MsgType.REQUEST:
            msg.request = self.request
            if self.request is Request.CONFIGURE:
                msg.config = protobuf.create_message(PROTO_CONFIG)
                msg.config.CopyFrom(self.config)
        return msg
    def get_keys(self) -> Iterable:
        """Returns iterable of dictionary keys to be used with `Protocol.handlers`.
        Keys must be provided in order of precedence (from more specific to general).
        """
        return [self.msg_type, ANY]

class _ICCP(Protocol):
    """Internal Component Control Protocol (ICCP).

    Used by Saturnin internally for component/controller transmissions.
    """
    #: string with protocol OID (dot notation).
    OID: str = '1.3.6.1.4.1.53446.1.2.0.1.1'
    # iso.org.dod.internet.private.enterprise.firebird.butler.platform.saturnin.protocol.iscp
    #: UUID instance that identifies the protocol.
    UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, OID)
    def __init__(self, *, session_type: Type[Session] = Session):
        """
        Arguments:
            session_type: Class for session objects.
        """
        super().__init__(session_type=session_type)
        self._msg: ICCPMessage = ICCPMessage()
        self.message_factory = self.__message_factory
    def __message_factory(self, zmsg: TZMQMessage = None) -> Message:
        "Internal message factory"
        self._msg.clear()
        return self._msg
    def validate(self, zmsg: TZMQMessage) -> None:
        """Verifies that sequence of ZMQ data frames is a valid protocol message.

        If this validation passes without exception, then `.parse()` of the same message
        must be successful as well.

        Arguments:
            zmsg:   ZeroMQ multipart message.

        Raises:
            InvalidMessageError: If ZMQ message is not a valid protocol message.
        """
        if not zmsg:
            raise InvalidMessageError("Empty message")
        try:
            msg_type = MsgType(unpack('!H', zmsg[0]))
        except Exception as exc:
            raise InvalidMessageError("Invalid message type") from exc
        if msg_type is MsgType.READY:
            try:
                PeerDescriptor.from_proto(protobuf.create_message(PROTO_PEER, zmsg[1]))
            except Exception as exc:
                raise InvalidMessageError("Invalid data: peer descriptor") from exc
            try:
                protobuf.create_message(protobuf.PROTO_STRUCT, zmsg[2])
            except Exception as exc:
                raise InvalidMessageError("Invalid data: endpoints") from exc
        elif msg_type is MsgType.ERROR:
            try:
                zmsg[1].decode('utf8')
            except Exception as exc:
                raise InvalidMessageError("Invalid data: error message") from exc
        elif msg_type is MsgType.REQUEST:
            try:
                req = Request(zmsg[1])
            except Exception as exc:
                raise InvalidMessageError("Invalid request code") from exc
            if req is Request.CONFIGURE:
                try:
                    protobuf.create_message(PROTO_CONFIG, zmsg[2])
                except Exception as exc:
                    raise InvalidMessageError("Invalid data: config") from exc
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <class_name>"
        return getattr(self, '_logging_id_', self.__class__.__name__)

class ICCPComponent(_ICCP):
    """Internal Component Control Protocol (ICCP) - Component (client) side.

    Used by Saturnin internally for component/controller transmissions.
    """
    def __init__(self, *, session_type: Type[Session] = Session, with_traceback: bool=False):
        """
        Arguments:
            session_type: Class for session objects.
            with_traceback: When True, stores traceback along with exception in ERROR/FINISHED.
        """
        super().__init__(session_type=session_type)
        self.with_traceback: bool = with_traceback
        self.handlers.update({MsgType.READY: self.wrong_message,
                              MsgType.REQUEST: self.handle_request,
                              MsgType.OK: self.wrong_message,
                              MsgType.ERROR: self.wrong_message,
                              MsgType.STOP: self.handle_stop,
                              MsgType.FINISHED: self.wrong_message,
                              })
    def accept_new_session(self, channel: Channel, routing_id: RoutingID,
                           msg: ICCPMessage) -> bool:
        """Validates incoming message that initiated new session/transmission.

        Arguments:
            channel:    Channel that received the message.
            routing_id: Routing ID of the sender.
            msg:        Received message.

        Returns:
            Always False (transmission must be initiated by Component).
        """
        return False
    def connect_with_session(self, channel: Channel) -> bool:
        """Called by :meth:`Channel.connect` to determine whether new session should be
        associated with connected peer.

        As ICCP require that connecting peers must send a message to initiate transmission,
        it always returns True.
        """
        return True
    def wrong_message(self, channel: Channel, session: Session, msg: ICCPMessage) -> None:
        """Handle wrong message received from controller.

        Raises `StopError`, which in turn calls `on_exception` from `handle_msg`.
        """
        raise StopError("Wrong message received from controller")
    def handle_invalid_msg(self, channel: Channel, session: Session, exc: InvalidMessageError) -> None:
        """Event handler for `on_invalid_msg`. Calls `on_stop_component` with exception.
        """
        self.on_stop_component(exc)
        super().handle_invalid_msg(channel, session, exc)
    def handle_exception(self, channel: Channel, session: Session, msg: ICCPMessage, exc: Exception) -> None:
        """Event handler for `on_exception`. Calls `on_stop_component` with exception.
        """
        self.on_stop_component(exc)
        super().handle_exception(channel, session, msg, exc)
    def handle_stop(self, channel: Channel, session: Session, msg: ICCPMessage) -> None:
        """Process STOP message received from controller. Calls `on_stop_component`.
        """
        self.on_stop_component()
    def handle_request(self, channel: Channel, session: Session, msg: ICCPMessage) -> None:
        """Process REQUEST message received from controller.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        result = self.ok_msg()
        try:
            if msg.data is Request.CONFIGURE:
                self.on_config_request(msg.config)
        except Exception as exc:
            result = self.error_msg(exc)
        if not (err_code := channel.send(result, session)):
            raise StopError("Send to controller failed", err_code=err_code)
    def ready_msg(self, peer: PeerDescriptor,
                  endpoints: Dict[str, List[ZMQAddress]]) -> ICCPMessage:
        """Returns READY control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.READY
        msg.peer = peer.copy()
        msg.endpoints = endpoints.copy()
        return msg
    def ok_msg(self) -> ICCPMessage:
        """Returns OK control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.OK
        return msg
    def error_msg(self, exc: Exception) -> ICCPMessage:
        """Returns ERROR control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.ERROR
        if self.with_traceback:
            msg.error = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__))
        else:
            msg.error = str(exc)
        return msg
    def finished_msg(self, outcome: Outcome, details: Union[None, Exception, List[str]]) -> ICCPMessage:
        """Returns FINISHED control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.FINISHED
        msg.outcome = outcome
        if isinstance(details, Exception):
            if self.with_traceback:
                msg.details = traceback.format_exception(type(details), details, details.__traceback__)
            else:
                msg.details = repr(details)
        elif details is None:
            msg.details = []
        else:
            msg.details = details.copy()
        return msg
    @eventsocket
    def on_stop_component(self, exc: Exception=None) -> None:
        """Called when commponent should stop its operation.

        Arguments:
           exc: Exception that describes the reason why component should stop. If not
                provided, the component should stop on controller's request.
        """
    @eventsocket
    def on_config_request(self, config: ConfigProto) -> None:
        """Called when controller requested reconfiguration.

        Any exception raised by event handler is returned back to controller via ERROR
        message.

        Arguments:
           config: New configuration provided by controller.
        """

class ICCPController(_ICCP):
    """Internal Component Control Protocol (ICCP) - Controller (server) side.

    Used by Saturnin internally for component/controller transmissions.
    """
    def __init__(self, *, session_type: Type[Session] = Session):
        """
        Arguments:
            session_type: Class for session objects.
        """
        super().__init__(session_type=session_type)
        self.handlers.update({MsgType.READY: self.handle_ready,
                              MsgType.REQUEST: self.wrong_message,
                              MsgType.OK: self.handle_oef,
                              MsgType.ERROR: self.handle_oef,
                              MsgType.STOP: self.wrong_message,
                              MsgType.FINISHED: self.handle_oef,
                              })
    def wrong_message(self, channel: Channel, session: Session, msg: ICCPMessage) -> None:
        """Handle wrong message received from component.

        Raises `StopError`, which in turn calls `on_exception` from `handle_msg`.
        """
        raise StopError("Wrong message received from component")
    def handle_invalid_msg(self, channel: Channel, session: Session, exc: InvalidMessageError) -> None:
        """Event handler for `on_invalid_msg`. Calls `on_stop_controller` with exception.
        """
        self.on_stop_controller(exc)
        super().handle_invalid_msg(channel, session, exc)
    def handle_exception(self, channel: Channel, session: Session, msg: ICCPMessage, exc: Exception) -> None:
        """Event handler for `on_exception`. Calls `on_stop_controller` with exception.
        """
        self.on_stop_controller(exc)
        super().handle_exception(channel, session, msg, exc)
    def handle_ready(self, channel: Channel, session: Session, msg: ICCPMessage) -> ICCPMessage:
        """Process READY message received from component.

        Returns:
            Received READY message if it's the first one from component.

        Raises:
            StopError: If it's NOT first READY received from component.
        """
        if hasattr(session, 'ready'):
            raise StopError("Unexpected READY message from component")
        session.ready = True
        return msg
    def handle_oef(self, channel: Channel, session: Session, msg: ICCPMessage) -> ICCPMessage:
        """Process OK/ERROR/FINISHED messages received from component. It simply returns
        the message.
        """
        if msg.msg_type in [MsgType.OK, MsgType.FINISHED] and not hasattr(session, 'ready'):
            raise StopError(f"Unexpected {msg.msg_type.name} message from component")
        return msg
    def stop_msg(self) -> ICCPMessage:
        """Returns OK control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.STOP
        return msg
    def request_config_msg(self, config: Config=None) -> ICCPMessage:
        """Returns REQUEST/CONFIG control message.
        """
        msg: ICCPMessage = self.message_factory()
        msg.msg_type = MsgType.REQUEST
        msg.request = Request.CONFIGURE
        msg.config = protobuf.create_message(PROTO_CONFIG)
        if config:
            config.save_proto(msg.config)
        return msg
    @eventsocket
    def on_stop_controller(self, exc: Exception) -> None:
        """Called when controller should stop its operation due to error condition.

        Arguments:
           exc: Exception that describes the reason why component should stop.
        """



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/micro.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Microservices
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin base module for implementation of Firebird Butler Microservices
"""

from __future__ import annotations
from typing import Union, Dict, List, cast
import sys
import os
import platform
import threading
import uuid
import zmq
from weakref import proxy
from time import monotonic_ns
from heapq import heappush, heappop
from firebird.base.types import Conjunctive
from firebird.base.trace import TracedMixin
from saturnin.base import ZMQAddress, Component, PeerDescriptor, ServiceDescriptor, \
     ServiceError, Direction, State, Outcome, ChannelManager, Channel, PairChannel, \
     ComponentConfig, ConfigProto, PrioritizedItem
from saturnin.protocol.iccp import ICCPComponent

#: Service control channel name
SVC_CTRL = sys.intern('iccp')

class MicroService(Component, TracedMixin, metaclass=Conjunctive):
    """Saturnin Component for Firebird Burler Microservices.
    """
    def __init__(self, zmq_context: zmq.Context, descriptor: ServiceDescriptor, *,
                 peer_uid: uuid.UUID=None):
        """
        Arguments:
            zmq_context: ZeroMQ Context.
            descriptor: Service descriptor.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
        """
        self._heap: List = []
        #: Service execution outcome
        self.outcome: Outcome = Outcome.UNKNOWN
        #: Service execution outcome details
        self.details: Union[Exception, List[str]] = None
        #: Service internal state
        self.state: State = State.UNKNOWN_STATE
        #: Event to stop the component
        self.stop: threading.Event = threading.Event()
        #: ChannelManager instance.
        self.mngr: ChannelManager = ChannelManager(zmq_context)
        self.mngr.log_context = proxy(self)
        #: Dictionary with endpoints to which the component binds.
        #: Key is channel name, value is list of ZMQAddress instances.
        #: Initially empty.
        self.endpoints: Dict[str, List[ZMQAddress]] = {}
        #: Service desriptor.
        self.descriptor: ServiceDescriptor = descriptor
        #: Peer descriptor for this component.
        self.peer: PeerDescriptor = PeerDescriptor(uuid.uuid1() if peer_uid is None else peer_uid,
                                                   os.getpid(), platform.node())
    def __str__(self):
        return self.logging_id
    __repr__ = __str__
    def handle_stop_component(self, exc: Exception=None) -> None:
        """ICCP event handler. Called when commponent should stop its operation.
        It stops the component by setting the `~Component.stop` event.

        Arguments:
           exc: Exception that describes the reason why component should stop. If not
                provided, the component should stop on controller's request.
        """
        if exc is not None:
            self.outcome = Outcome.ERROR
            self.details = exc
        self.stop.set()
    def handle_config_request(self, config: ConfigProto) -> None:
        """ICCP event handler. Called when controller requested reconfiguration.

        Must raise an exception if configuration fails for any reason.

        By default, the component does not support run-time configuration, so it raises
        NotImplementedError.

        Arguments:
           config: New configuration provided by controller.
        """
        raise NotImplementedError("Service does not support run-time configuration")
    def schedule(self, action: Callable, after: int) -> None:
        """Schedule action to be executed after specified time.

        Action is executed in `run()` main I/O loop not sooner than after specified number
        of milliseconds from time when `schedule()` is called. However, the actual delay
        could be longer than specified (depends on time spent in message handlers and other
        factors).

        Arguments:
            action: Callable (without arguments) to be executed. Use `functools.partial`
                    if callable requires arguments.
            after:  Delay in milliseconds.
        """
        heappush(self._heap, PrioritizedItem(monotonic_ns() + (after * 1000000), action))
    def get_timeout(self) -> int:
        """Returns timeout to next scheduled action.
        """
        if not self._heap:
            return None
        back = []
        i = len(self._heap)
        now = monotonic_ns()
        while self._heap and (item := heappop(self._heap)).priority < now:
            back.append(item)
        if len(back) != i:
            heappush(self._heap, item)
        for b in back:
            heappush(self._heap, b)
        return max(int((item.priority - now) / 1000000), 0)
    def run_scheduled(self) -> None:
        """Run scheduled actions.
        """
        while self._heap:
            item = heappop(self._heap)
            if item.priority < monotonic_ns():
                item.item()
            else:
                heappush(self._heap, item)
                break
    def initialize(self, config: ComponentConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        config.validate() # Fail early!
        if config.logging_id.value is not None:
            self._logging_id_ = config.logging_id.value
    def bind_endpoints(self) -> None:
        """Bind endpoints used by component.
        """
        for name, addr_list in self.endpoints.items():
            chn: Channel = self.mngr.channels.get(name)
            for i in range(len(addr_list)):
                self.endpoints[name][i] = chn.bind(addr_list[i])
    def aquire_resources(self) -> None:
        """Aquire resources required by component (open files, connect to other services etc.).

        Must raise an exception when resource aquisition fails.
        """
    def release_resources(self) -> None:
        """Release resources aquired by component (close files, disconnect from other services etc.)
        """
    def start_activities(self) -> None:
        """Start normal component activities.

        Must raise an exception when start fails.
        """
    def stop_activities(self) -> None:
        """Stop component activities.
        """
    def warm_up(self, ctrl_addr: Optional[ZMQAddress]) -> None:
        """Initializes the ChannelManager and connects component to control channel.
        """
        if ctrl_addr is not None:
            # Service control channel
            iccp = ICCPComponent(with_traceback=__debug__)
            iccp.on_stop_component = self.handle_stop_component
            iccp.on_config_request = self.handle_config_request
            chn: PairChannel = self.mngr.create_channel(PairChannel, SVC_CTRL, iccp,
                                                        wait_for=Direction.IN,
                                                        sock_opts={'rcvhwm': 5,
                                                                   'sndhwm': 5,})
            chn.protocol.log_context = self.logging_id
        self.mngr.warm_up()
        if ctrl_addr is not None:
            chn.connect(ctrl_addr)
            if not chn.can_send():
                raise ServiceError("Broken component control channel")
        try:
            self.bind_endpoints()
            self.aquire_resources()
            self.start_activities()
        except Exception as exc:
            if ctrl_addr is not None:
                chn.send(cast(ICCPComponent, chn.protocol).error_msg(exc), chn.session)
            self.mngr.shutdown()
            raise
        else:
            if ctrl_addr is not None:
                chn.send(cast(ICCPComponent, chn.protocol).ready_msg(self.peer, self.endpoints),
                         chn.session)
            self.state = State.READY
    def run(self) -> None:
        """Component execution (main loop).
        """
        self.state = State.RUNNING
        ctrl_chn: PairChannel = self.mngr.channels.get(SVC_CTRL)
        try:
            while not self.stop.is_set():
                events = self.mngr.wait(self.get_timeout())
                if events:
                    # Messages from service control channel have top priority
                    if ctrl_chn in events:
                        ctrl_chn.receive()
                        if self.stop.is_set():
                            continue # stop quickly
                    # Channels waiting for output have precedence
                    if self.mngr.has_pollout():
                        for chn, event in events.items():
                            if Direction.OUT in event:
                                chn.on_output_ready(chn)
                    # Now process incomming messages
                    for chn, event in events.items():
                        if Direction.IN in event:
                            chn.receive()
                # Now it's time for scheduled actions
                self.run_scheduled()
            # Gracefully stop the service
            self.state = State.STOPPED
            self.stop_activities()
            self.release_resources()
            if self.outcome is Outcome.UNKNOWN:
                self.outcome = Outcome.OK
            ctrl_chn.send(cast(ICCPComponent, ctrl_chn.protocol).finished_msg(self.outcome,
                                                                              self.details),
                          ctrl_chn.session)
            self.mngr.shutdown()
            self.state = State.FINISHED
        except Exception as exc:
            self.state = State.ABORTED
            try:
                # try send report to controller
                ctrl_chn.send(cast(ICCPComponent, ctrl_chn.protocol).error_msg(exc),
                              ctrl_chn.session)
            except:
                pass
            try:
                self.mngr.shutdown(forced=True)
            except:
                pass
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <agent_name>[<peer.uid.hex>]"
        return getattr(self, '_logging_id_', f'{self.descriptor.agent.name}[{self.peer.uid.hex}]')




#coding:utf-8
#
# PROGRAM/MODULE: saturnin-core
# FILE:           saturnin/core/types.py
# DESCRIPTION:    Type definitions
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - Type definitions
"""

from __future__ import annotations
from typing import List, Dict, Optional, NewType, Callable, Any
from enum import IntEnum, IntFlag, Enum, auto
import uuid
import zmq
from dataclasses import dataclass, field, replace as _dcls_replace
from firebird.base.types import Error, Distinct, MIME, Sentinel
from firebird.base.protobuf import PROTO_STRUCT, load_registered, create_message, \
     struct2dict, dict2struct

# Type annotation types
TSupplement = Optional[Dict[str, Any]]
"""name/value dictionary"""
Token = NewType('Token', bytearray)
"""Message token"""
RoutingID = NewType('RoutingID', bytes)
"""Routing ID"""

# Constants
PLATFORM_OID: str = '1.3.6.1.4.1.53446.1.2.0'
"Platform OID (`firebird.butler.platform.saturnin`)"
PLATFORM_UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, PLATFORM_OID)
"Platform UID (:func:`~uuid.uuid5` - NAMESPACE_OID)"
PLATFORM_VERSION: str = '0.6.0'
"Platform version (semver)"

VENDOR_OID: str = '1.3.6.1.4.1.53446.1.3.0'
"Platform vendor OID (`firebird.butler.vendor.firebird`)"
VENDOR_UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, VENDOR_OID)
"Platform vendor UID (:func:`~uuid.uuid5` - NAMESPACE_OID)"

MIME_TYPE_PROTO = MIME('application/x.fb.proto')
MIME_TYPE_TEXT = MIME('text/plain')
MIME_TYPE_BINARY = MIME('application/octet-stream')

PROTO_PEER = 'firebird.butler.PeerIdentification'

#  Exceptions
class InvalidMessageError(Error):
    "A formal error was detected in a message"
class ChannelError(Error):
    "Transmission channel error"
class ServiceError(Error):
    "Error raised by service"
class ClientError(Error):
    "Error raised by Client"
class StopError(Error):
    "Error that should stop furter processing."

#Sentinels
#: Sentinel for return values that indicates failed message processing
INVALID: Sentinel = Sentinel('INVALID')
#: Sentinel for return values that indicates timeout expiration
TIMEOUT: Sentinel = Sentinel('TIMEOUT')

# Enums
class Origin(IntEnum):
    """Origin of received message in protocol context."""
    UNKNOWN = auto()
    SERVICE = auto()
    CLIENT = auto()
    ANY = auto()
    # Aliases
    PROVIDER = SERVICE
    CONSUMER = CLIENT
    def peer_role(self) -> Origin:
        if self == Origin.ANY:
            return self
        else:
            return Origin.CLIENT if self == Origin.SERVICE else Origin.SERVICE

class SocketMode(IntEnum):
    """ZeroMQ socket mode."""
    UNKNOWN = auto()
    BIND = auto()
    CONNECT = auto()

class Direction(IntFlag):
    """ZeroMQ socket direction of transmission."""
    NONE = 0
    IN = zmq.POLLIN
    OUT = zmq.POLLOUT
    BOTH = OUT | IN

class SocketType(IntEnum):
    """ZeroMQ socket type."""
    UNKNOWN_TYPE = -1 # Not a valid option, defined only to handle undefined values
    DEALER = zmq.DEALER
    ROUTER = zmq.ROUTER
    PUB = zmq.PUB
    SUB = zmq.SUB
    XPUB = zmq.XPUB
    XSUB = zmq.XSUB
    PUSH = zmq.PUSH
    PULL = zmq.PULL
    STREAM = zmq.STREAM
    PAIR = zmq.PAIR

class State(IntEnum):
    """General state information."""
    UNKNOWN_STATE = 0
    READY = 1
    RUNNING = 2
    WAITING = 3
    SUSPENDED = 4
    FINISHED = 5
    ABORTED = 6
    # Aliases
    CREATED = READY
    BLOCKED = WAITING
    STOPPED = SUSPENDED
    TERMINATED = ABORTED

class PipeSocket(IntEnum):
    """Data Pipe Socket identification."""
    UNKNOWN_PIPE_SOCKET = 0 # Not a valid option, defined only to handle undefined values
    INPUT = 1
    OUTPUT = 2

class FileOpenMode(IntEnum):
    """File open mode."""
    UNKNOWN_FILE_OPEN_MODE = 0 # Not a valid option, defined only to handle undefined values
    READ = 1
    CREATE = 2
    WRITE = 3
    APPEND = 4
    RENAME = 5

class Outcome(Enum):
    """Service execution outcome.
    """
    UNKNOWN = 'UNKNOWN'
    OK = 'OK'
    ERROR = 'ERROR'

class ButlerInterface(IntEnum):
    """Base class for service API code enumerations (FBSP interfaces).
    """
    @classmethod
    def get_uid(cls) -> uuid.UUID:
        raise NotImplementedError()

# Dataclasses
@dataclass(eq=True, order=False, frozen=True)
class AgentDescriptor(Distinct):
    """Service or Client descriptor dataclass.

    Note:
        Because this is a `dataclass`, the class variables are those attributes that have
        default value. Other attributes are created in constructor.
    """
    #: Agent ID
    uid: uuid.UUID
    #: Agent name
    name: str
    #: Agent version string
    version: str
    #: Vendor ID
    vendor_uid: uuid.UUID
    #: Agent classification string
    classification: str
    #: Butler platform ID
    platform_uid: uuid.UUID = PLATFORM_UID
    #: Butler platform version string
    platform_version: str = PLATFORM_VERSION
    #: Optional list of supplemental information
    supplement: TSupplement = None
    def get_key(self) -> Any:
        """Returns `uid` (instance key). Used for instance hash computation."""
        return self.uid
    def copy(self) -> AgentDescriptor:
        """Returns copy of this AgentDescriptor instance.
        """
        return _dcls_replace(self)
    def replace(self, **changes) -> AgentDescriptor:
        """Creates a new `AgentDescriptor`, replacing fields with values from `changes`.
        """
        return _dcls_replace(self, **changes)

@dataclass(eq=True, order=False, frozen=True)
class PeerDescriptor(Distinct):
    """Peer descriptor.
    """
    #: Peer ID
    uid: uuid.UUID
    #: Peer process ID
    pid: int
    #: Host name
    host: str
    #: Optional list of supplemental information
    supplement: TSupplement = None
    def get_key(self) -> Any:
        """Returns `uid` (instance key). Used for instance hash computation."""
        return self.uid
    def as_proto(self) -> Any:
        """Returns `firebird.butler.PeerIdentification` protobuf message initialized
        from instance data.
        """
        msg = create_message(PROTO_PEER)
        msg.uid = self.uid.bytes
        msg.pid = self.pid
        msg.host = self.host
        if self.supplement is not None:
            sup = msg.supplement.add()
            sup.Pack(dict2struct(self.supplement))
        return msg
    def copy(self) -> PeerDescriptor:
        """Returns copy of this PeerDescriptor instance.
        """
        return _dcls_replace(self)
    def replace(self, **changes) -> PeerDescriptor:
        """Creates a new `PeerDescriptor`, replacing fields with values from `changes`.
        """
        return _dcls_replace(self, **changes)
    @classmethod
    def from_proto(cls, proto: Any) -> PeerDescriptor:
        """Creates new PeerDescriptor from `firebird.butler.PeerIdentification` protobuf
        message.
        """
        if proto.DESCRIPTOR.full_name != PROTO_PEER:
            raise ValueError("PeerIdentification protobuf message required")
        data = None
        if proto.supplement:
            for i in proto.supplement:
                if i.TypeName() == PROTO_STRUCT:
                    s = create_message(PROTO_STRUCT)
                    i.Unpack(s)
                    data = struct2dict(s)
                    break
        return cls(uuid.UUID(bytes=proto.uid), proto.pid, proto.host, data)

@dataclass(eq=True, order=False, frozen=True)
class ServiceDescriptor(Distinct):
    """Service descriptor.
    """
    #: Service agent descriptor
    agent: AgentDescriptor
    #: Service FBSP API description or `None` for microservice
    api: List[ButlerInterface]
    #: Text describing the service
    description: str
    #: List of Saturnin facilities that this service uses
    facilities: List[str]
    #: Python package that contains this service
    package: str
    #: Locator string for service factory
    factory: str
    #: Service configuration factory
    config: Callable[[], 'firebird.base.config.Config']
    def get_key(self) -> Any:
        """Returns `agent.uid` (instance key). Used for instance hash computation."""
        return self.agent.uid

@dataclass(order=True)
class PrioritizedItem:
    """Prioritized item for use with `heapq` to implement priority queue.
    """
    priority: int
    item: Any=field(compare=False)

load_registered('firebird.butler.protobuf')
load_registered('firebird.base.protobuf')
del load_registered



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/bundlerun.py
# DESCRIPTION:    Script to run bundle of services
# CREATED:        6.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin script to run bundle of services


"""

from __future__ import annotations
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, Action
from configparser import ConfigParser, ExtendedInterpolation
import logging
from logging.config import fileConfig
from datetime import datetime
import zmq
from firebird.base.logging import get_logger, Logger
from firebird.base.trace import trace_manager
from saturnin.base import ChannelManager
from saturnin.component.bundle import BundleThreadController
from saturnin.component.controller import Outcome

LOG_FORMAT = '%(levelname)s [%(processName)s/%(threadName)s] [%(agent)s:%(context)s] %(message)s'

class UpperAction(Action):
    """Converts argument to uppercase.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.upper())

PROG_NAME = 'saturnin-bundle'

def main():
    """Runs bundle of services.
    """
    parser: ArgumentParser = ArgumentParser(PROG_NAME, description=main.__doc__,
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('service', metavar='SERVICE-CONFIG',
                        help="Path to service configuration file")
    parser.add_argument('-c','--config', metavar='CONFIG', action='append',
                        help="Path to additional configuration file. Could be specified multiple times.")
    parser.add_argument('-s', '--section', help="Configuration section name", default='bundle')
    parser.add_argument('-o','--outcome', action='store_true',
                        help="Always print service execution outcome", default=False)
    parser.add_argument('-l', '--log-level', action=UpperAction,
                        choices=[x.lower() for x in logging._nameToLevel
                                 if isinstance(x, str)],
                        help="Logging level")

    args = parser.parse_args()

    main_config: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())

    mngr: ChannelManager = None
    log: Logger = None
    try:
        cfg_files = []
        if args.config:
            cfg_files.extend(args.config)
        cfg_files.append(args.service)
        cfg_files = main_config.read(cfg_files)
        # Logging configuration
        if main_config.has_section('loggers'):
            fileConfig(main_config)
        else:
            logging.basicConfig(format=LOG_FORMAT)
        log = get_logger(PROG_NAME)
        if args.log_level is not None:
            log.setLevel(args.log_level)
        # trace configuration
        if main_config.has_section('trace'):
            trace_manager.load_config(main_config)
        #
        mngr: ChannelManager = ChannelManager(zmq.Context.instance())
        mngr.log_context = PROG_NAME
        executor: BundleThreadController = BundleThreadController(manager=mngr)
        executor.log_context = PROG_NAME
        executor.config.read(cfg_files)
        executor.configure(section=args.section)
        # run services
        start = datetime.now()
        executor.start()
        try:
            executor.join()
            raise KeyboardInterrupt() # This, or direct call to executor.stop()
        except KeyboardInterrupt: # SIGINT
            executor.stop()
        finally:
            for svc in executor.services:
                if svc.outcome is not Outcome.OK or args.outcome:
                    print(f'{svc.name} outcome:', svc.outcome.value)
                    if svc.details:
                        for line in svc.details:
                            print(' ', line)
            print('Execution time:', datetime.now() - start)
    except Exception as exc:
        if log:
            log.exception("Service execution failed")
        parser.exit(1, f'{exc!s}\n')
    finally:
        if mngr is not None:
            mngr.shutdown(forced=True)
        logging.shutdown()
        zmq.Context.instance().term()

if __name__ == '__main__':
    main()





#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/svcrun.py
# DESCRIPTION:    Script to run one service in main or separate thread
# CREATED:        19.11.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin script to run one service in main or separate thread

"""

from __future__ import annotations
import sys
from os import getcwd
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, Action
from configparser import ConfigParser, ExtendedInterpolation, DEFAULTSECT
import logging
from logging.config import fileConfig
import zmq
from firebird.base.logging import get_logger, Logger
from firebird.base.trace import trace_manager
from saturnin.base import StopError, ServiceDescriptor
from saturnin.component.registry import get_service_desciptors
from saturnin.component.controller import Outcome, DirectController, ThreadController, \
     ServiceExecConfig

LOG_FORMAT = '%(levelname)s [%(processName)s/%(threadName)s] [%(agent)s:%(context)s] %(message)s'

class UpperAction(Action):
    """Converts argument to uppercase.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.upper())

PROG_NAME = 'saturnin-service'

def main():
    """Runs one service, either unmanaged in main thread, or managed in separate thread.
    """
    parser: ArgumentParser = ArgumentParser(PROG_NAME, description=main.__doc__,
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('service', metavar='SERVICE-CONFIG',
                        help="Path to service configuration file")
    parser.add_argument('-c','--config', metavar='CONFIG', action='append',
                        help="Path to additional configuration file. Could be specified multiple times.")
    parser.add_argument('-s', '--section', help="Configuration section name", default='service')
    parser.add_argument('-o','--outcome', action='store_true',
                        help="Always print service execution outcome", default=False)
    parser.add_argument('--main-thread', action='store_true',
                        help="Start the service in main thread", default=False)
    parser.add_argument('-l', '--log-level', action=UpperAction,
                        choices=[x.lower() for x in logging._nameToLevel
                                 if isinstance(x, str)],
                        help="Logging level")

    args = parser.parse_args()

    main_config: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
    # Defaults
    main_config[DEFAULTSECT]['here'] = getcwd()

    #: Could be used to stop the service in debugger session
    debug_stop: bool = False
    log: Logger = None
    try:
        # Read config
        cfg_files = []
        if args.config:
            cfg_files.extend(args.config)
        cfg_files.append(args.service)
        cfg_files = main_config.read(cfg_files)
        # Logging configuration
        if main_config.has_section('loggers'):
            fileConfig(main_config)
        else:
            logging.basicConfig(format=LOG_FORMAT)
        log = get_logger(PROG_NAME)
        if args.log_level is not None:
            log.setLevel(args.log_level.upper())
        # trace configuration
        if main_config.has_section('trace'):
            trace_manager.load_config(main_config)
        #
        if not main_config.has_section(args.section):
            raise StopError(f"Missing configuration section '{args.section}'")
        svc_executor_config: ServiceExecConfig = ServiceExecConfig(args.section)
        svc_executor_config.load_config(main_config)
        entries = get_service_desciptors(str(svc_executor_config.agent.value))
        if not entries:
            raise StopError(f"Unregistered agent '{svc_executor_config.agent.value}'")
        service_desc: ServiceDescriptor = entries[0]
        #
        if args.main_thread:
            executor: DirectController = DirectController(service_desc)
        else:
            executor: ThreadController = ThreadController(service_desc)
        executor.log_context = PROG_NAME
        executor.configure(main_config, args.section)
        # run the service
        executor.start(timeout=10000 if sys.gettrace() is None else None)
        if not args.main_thread:
            try:
                while True:
                    executor.join(1)
                    # This, or direct call to executor.stop()
                    if debug_stop or not executor.is_running():
                        raise KeyboardInterrupt()
            except KeyboardInterrupt: # SIGINT
                print()
                try:
                    executor.stop()
                except TimeoutError:
                    executor.terminate()
                except Exception as exc:
                    if log:
                        log.error("Error while stopping the service")
            finally:
                if executor.outcome is not Outcome.OK or args.outcome:
                    print('Outcome:', executor.outcome.value)
                    if executor.details:
                        print('Details:')
                        for line in executor.details:
                            print(line)
    except Exception as exc:
        if log:
            log.error("Service execution failed")
        parser.exit(1, f'{exc!s}\n')
    finally:
        logging.shutdown()
        zmq.Context.instance().term()

if __name__ == '__main__':
    main()




#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/daemon.py
# DESCRIPTION:    Script to start/stop the daemon process.
# CREATED:        9.11.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin script to start/stop the daemon process.


"""

from __future__ import annotations
import platform
import argparse
from pathlib import Path
from saturnin.lib.daemon import start_daemon

def main():
    """Starts or stops the daemon process.
    """
    parser = argparse.ArgumentParser('saturnin-daemon', description=main.__doc__)
    subs = parser.add_subparsers(title="Commands", dest='action', required=True)
    start_args = subs.add_parser('start', help="Starts daemon process")
    start_args.add_argument('pid_file', help="Path to PID file")
    start_args.add_argument('daemon', help="Path to daemon")
    start_args.add_argument('arguments', nargs=argparse.REMAINDER, help="Daemon arguments")

    stop_args = subs.add_parser('stop', help="Stops daemon process")
    stop_args.add_argument('pid', help="Daemon PID or path to PID file")

    args = parser.parse_args()

    if args.action == 'start':
        pid_file = Path(args.pid_file)
        pid_file.write_text('') # ensure it's writtable
        args.arguments.insert(0, args.daemon)
        pid = start_daemon(args.arguments)
        pid_file.write_text(str(pid))
    else: # stop
        try:
            pid = int(args.pid)
        except ValueError:
            pid = int(Path(args.pid).read_text())
        if platform.system() == 'Windows':
            import ctypes
            kernel = ctypes.windll.kernel32
            kernel.FreeConsole()
            kernel.AttachConsole(pid)
            kernel.SetConsoleCtrlHandler(None, 1)
            kernel.GenerateConsoleCtrlEvent(0, 0)
        else:  # Unix
            import os
            import signal
            os.kill(pid, signal.SIGINT)

if __name__ == '__main__':
    main()



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/daemon.py
# DESCRIPTION:    Daemon process management
# CREATED:        6.11.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin daemon process management


"""

from __future__ import annotations
from typing import Union
import platform
import subprocess
from pathlib import Path

def start_daemon(args: list) -> int:
    """Starts daemon process.

    Arguments:
        args: Arguments for `subprocess.Popen` (first item must be the daemon filename)

    Returns:
        PID for started daemon, or None if start failed.

    Note:
        Gracefull shutdown on Windows is tricky. To allow shutdown via SIGINT signal handler
        (note that SIGINT is not available on Windows, you have to use CTRL_C_EVENT),
        it's necessry to start new shell with new console in background (i.e. detached from
        console of this daemon starting script).
    """
    kwargs = {}
    if platform.system() == 'Windows':
        kwargs.update(shell=True, creationflags=subprocess.CREATE_NEW_CONSOLE)
    else:  # Unix
        kwargs.update(start_new_session=True)
    p = subprocess.Popen(args, **kwargs)
    return p.pid if p.poll() is None else None

def stop_daemon(pid: Union[int, str, Path]) -> None:
    """Stops the daemon process by invoking `saturnin-daemon` script.

    Arguments:
        pid: PID or text file name/Path where PID is stored.

    Raises:
        subprocess.CalledProcessError: When execution of `saturnin-daemon` failed.
        subprocess.TimeoutExpired: When `saturnin-daemon` does not finish in 5 seconds.

    Important:
        On Linux/Unix: Sends SIGINT signal to the daemon process.
        On Windows: Detaches from console, attaches itself to daemon console and sends
        control-C event to it.

    Note:
        Gracefull shutdown on Windows is tricky. It requires that the daemon process has
        a console, otherwise the CTRL_C_EVENT couldn't be delivered. This condition is met
        if daemon was started by :func:`start_daemon` or `saturnin-daemon` script.
    """
    if isinstance(pid, int):
        pid = str(pid)
    subprocess.run(['saturnin-daemon', 'stop', pid], check=True, timeout=5)



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/service.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Services
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin base module for implementation of Firebird Butler Services
"""

from __future__ import annotations
from typing import cast
from abc import abstractmethod
from firebird.base.config import ListOption
from saturnin.base import ZMQAddress, RouterChannel, ComponentConfig
from saturnin.component.micro import MicroService
from saturnin.protocol.fbsp import FBSPService

SVC_CHN = 'service'

class ServiceConfig(ComponentConfig):
    """Base data provider/consumer microservice configuration.
    """
    def __init__(self, name: str):
        super().__init__(name)
        self.endpoints: ListOption = \
            ListOption('endpoints', ZMQAddress, "List of service endpoints", required=True)

class Service(MicroService):
    """
    """
    def initialize(self, config: ServiceConfig) -> None:
        """Verify configuration and assemble service structural parts.
        """
        super().initialize(config)
        # Set endpoints this service binds
        self.endpoints[SVC_CHN] = config.endpoints.value.copy()
        #: Service protocol
        service = FBSPService(service=self.descriptor, peer=self.peer)
        service.log_context = self.logging_id
        #: Channel for communication with service clients.
        self.svc_channel = self.mngr.create_channel(RouterChannel, SVC_CHN, service,
                                                    routing_id=self.peer.uid.hex.encode('ascii'),
                                                    sock_opts={'maxmsgsize': 52428800,
                                                               'rcvhwm': 500,
                                                               'sndhwm': 500,})
        self.register_api_handlers(service)
    @abstractmethod
    def register_api_handlers(self, service: FBSPService) -> None:
        """Called by `initialize()` for registration of service API handlers and FBSP
        service event handlers.
        """
    def start_activities(self) -> None:
        """Start normal service activities.

        Must raise an exception when start fails.
        """
        super().start_activities()
        self.svc_channel.set_wait_in(True)
    def stop_activities(self) -> None:
        """Stop service activities.
        """
        super().stop_activities()
        cast(FBSPService, self.svc_channel.protocol).close(self.svc_channel)
        self.svc_channel.set_wait_in(False)



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/base.py
# DESCRIPTION:    Abstract base class for Saturnin Components
# CREATED:        3.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin abstract base class for Components

"""

from __future__ import annotations
from abc import ABC, abstractmethod
from firebird.base.types import ZMQAddress
from firebird.base.config import Config, StrOption

class ComponentConfig(Config):
    """Base Component configuration."""
    def __init__(self, name: str):
        super().__init__(name)
        self.logging_id: StrOption = \
            StrOption('logging_id', "Logging ID for this component instance")

class Component(ABC):
    """Abstract base class for Saturnin Components.
    """
    @abstractmethod
    def initialize(self, config: ComponentConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
    @abstractmethod
    def warm_up(self, ctrl_addr: Optional[ZMQAddress]) -> None:
        """Must initialize the ChannelManager and connect component to control channel
        at provided address.
        """
    @abstractmethod
    def run(self) -> None:
        """Component execution (main loop).
        """



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/protocol/fbdp.py
# DESCRIPTION:    Firebird Butler Data Pipe Protocol
#                 See https://firebird-butler.readthedocs.io/en/latest/rfc/9/FBDP.html
# CREATED:        30.7.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin reference implementation of Firebird Butler Data Pipe Protocol

See https://firebird-butler.readthedocs.io/en/latest/rfc/9/FBDP.html
"""

from __future__ import annotations
from typing import Type, Dict, Any, Union
import uuid
import warnings
from struct import pack, unpack
from enum import IntEnum, IntFlag
from firebird.base.signal import eventsocket
from firebird.base.protobuf import ProtoMessage, create_message, dict2struct, struct2dict
from saturnin.base import InvalidMessageError, StopError, RoutingID, TZMQMessage, \
     PipeSocket, Channel, Protocol, Message, Session, ANY

PROTO_OPEN = 'firebird.butler.FBDPOpenDataframe'
PROTO_ERROR = 'firebird.butler.ErrorDescription'

#: FBDP protocol control frame :mod:`struct` format
HEADER_FMT_FULL : str = '!4sBBH'
#: FBDP protocol control frame :mod:`struct` format without FOURCC
HEADER_FMT: str = '!4xBBH'
#: FBDP protocol identification (FOURCC)
FOURCC: bytes = b'FBDP'
#: FBDP protocol version mask
VERSION_MASK: int = 7

#: Default data batch size
DATA_BATCH_SIZE: int = 50

class MsgType(IntEnum):
    """FBDP Message Type"""
    UNKNOWN = 0 # not a valid message type
    OPEN = 1    # initial message from client
    READY = 2   # transfer negotiation message
    NOOP = 3    # no operation, used for keep-alive & ping purposes
    DATA = 4    # user data
    CLOSE = 5   # sent by peer that is going to close the connection

class MsgFlag(IntFlag):
    """FBDP message flag"""
    NONE = 0
    ACK_REQ = 1
    ACK_REPLY = 2

class ErrorCode(IntEnum):
    """FBDP Error Code"""
    # No error
    OK = 0
    # General errors
    INVALID_MESSAGE = 1
    PROTOCOL_VIOLATION = 2
    ERROR = 3
    INTERNAL_ERROR = 4
    INVALID_DATA = 5
    TIMEOUT = 6
    # Errors that prevent the connection from opening
    PIPE_ENDPOINT_UNAVAILABLE = 100
    FBDP_VERSION_NOT_SUPPORTED = 101
    NOT_IMPLEMENTED = 102
    DATA_FORMAT_NOT_SUPPORTED = 103


class FBDPMessage(Message):
    """Firebird Butler Datapipe Protocol (FBDP) Message.
    """
    def __init__(self):
        #: Type of message
        self.msg_type: MsgType = MsgType.UNKNOWN
        #: Message flags
        self.flags: MsgFlag = MsgFlag(0)
        #: Data associated with message
        self.type_data: int = 0
        #: Data frame associated with message type (or None)
        self.data_frame: Union[ProtoMessage, Any] = None
    def __str__(self):
        return f"{self.__class__.__qualname__}[{self.msg_type.name}]"
    __repr__ = __str__
    def from_zmsg(self, zmsg: TZMQMessage) -> None:
        """Populate message data from sequence of ZMQ data frames.

        Arguments:
            zmsg: Sequence of frames that should be deserialized.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.
        """
        try:
            control_byte, flags, self.type_data = unpack(HEADER_FMT, zmsg.pop(0))
            self.msg_type = MsgType(control_byte >> 3)
            self.flags = MsgFlag(flags)
            if self.msg_type is MsgType.OPEN:
                self.data_frame = create_message(PROTO_OPEN)
                self.data_frame.ParseFromString(zmsg.pop(0))
            elif self.msg_type is MsgType.DATA:
                self.data_frame = zmsg.pop(0) if zmsg else None
            elif self.msg_type is MsgType.CLOSE:
                self.type_data = ErrorCode(self.type_data)
                self.data_frame = []
                while zmsg:
                    err = create_message(PROTO_ERROR)
                    err.ParseFromString(zmsg.pop(0))
                    self.data_frame.append(err)
        except Exception as exc:
            raise InvalidMessageError("Invalid message") from exc
    def as_zmsg(self) -> TZMQMessage:
        """Returns message as sequence of ZMQ data frames.
        """
        zmsg = []
        zmsg.append(self.get_header())
        if self.msg_type is MsgType.OPEN:
            zmsg.append(self.data_frame.SerializeToString())
        elif (self.msg_type is MsgType.DATA and self.data_frame is not None):
            zmsg.append(self.data_frame)
        elif self.msg_type is MsgType.CLOSE:
            while self.data_frame:
                zmsg.append(self.data_frame.pop(0).SerializeToString())
        return zmsg
    def clear(self) -> None:
        """Clears message data.
        """
        self.msg_type = MsgType.UNKNOWN
        self.type_data = 0
        self.flags = MsgFlag(0)
        self.data_frame = None
    def copy(self) -> Message:
        """Returns copy of the message.
        """
        msg: FBDPMessage = self.__class__()
        msg.msg_type = self.msg_type
        msg.flags = self.flags
        msg.type_data = self.type_data
        if self.msg_type is MsgType.OPEN:
            msg.data_frame = create_message(PROTO_OPEN)
            msg.data_frame.CopyFrom(self.data_frame)
        elif self.msg_type is MsgType.CLOSE:
            msg.data_frame = []
            for frame in self.data_frame:
                err = create_message(PROTO_ERROR)
                err.CopyFrom(frame)
                msg.data_frame.append(err)
        else:
            msg.data_frame = self.data_frame
        return msg
    def get_keys(self) -> Iterable:
        """Returns iterable of dictionary keys to be used with `Protocol.handlers`.
        Keys must be provided in order of precedence (from more specific to general).
        """
        return [self.msg_type, ANY]
    def get_header(self) -> bytes:
        """Return message header (FBDP control frame).
        """
        return pack(HEADER_FMT_FULL, FOURCC, (self.msg_type << 3) | _FBDP.REVISION,
                    self.flags, self.type_data)
    def has_ack_req(self) -> bool:
        """Returns True if message has ACK_REQ flag set.
        """
        return MsgFlag.ACK_REQ in self.flags
    def has_ack_reply(self) -> bool:
        """Returns True if message has ASK_REPLY flag set.
        """
        return MsgFlag.ACK_REPLY in self.flags
    def set_flag(self, flag: MsgFlag) -> None:
        """Set flag specified by `flag` mask.
        """
        self.flags |= flag
    def clear_flag(self, flag: MsgFlag) -> None:
        """Clear flag specified by `flag` mask.
        """
        self.flags &= ~flag
    def note_exception(self, exc: Exception):
        """Store information from exception into CLOSE Message.
        """
        assert self.msg_type is MsgType.CLOSE
        errdesc = create_message(PROTO_ERROR)
        if hasattr(exc, 'code'):
            errdesc.code = getattr(exc, 'code')
        errdesc.description = str(exc)
        self.data_frame.append(errdesc)

class FBDPSession(Session):
    """FBDP session. Contains information about Data Pipe.
    """
    def __init__(self):
        super().__init__()
        #: Data Pipe Identification.
        self.pipe: str = None
        #: Data Pipe socket Identification.
        self.socket: PipeSocket = None
        #: Specification of format for user data transmitted in DATA messages.
        self.data_format: str = None
        #: Data Pipe parameters.
        self.params: Dict = {}
        #: Number of DATA messages that remain to be transmitted since last READY message.
        self.transmit: int = None
        #: Indicator that server sent READY and waits from READY response from client
        self.await_ready: bool = False

class _FBDP(Protocol):
    """9/FBDP - Firebird Butler Data Pipe Protocol
    """
    #: string with protocol OID (dot notation).
    OID: str =  '1.3.6.1.4.1.53446.1.5.2'
    # iso.org.dod.internet.private.enterprise.firebird.butler.protocol.fbdp
    #: UUID instance that identifies the protocol.
    UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, OID)
    def __init__(self, *, session_type: Type[FBDPSession] = FBDPSession):
        """
        Arguments:
            session_type: Class for session objects.
            batch_size: Default batch size.
        """
        super().__init__(session_type=session_type)
        #: Initial batch size
        self.batch_size = DATA_BATCH_SIZE
        #: CONSUMER option. Whether ACK_REPLY message for DATA/ACK_REQ should be sent
        #: before (False) or after (True) call to `.on_accept_data()` callback.
        self.confirm_processing: bool = False
        #: PRODUCER option. If sent DATA message has ACK_REQ flag set, send next DATA
        #: message after ACK_REPLY is received (True), or continue sending DATA without
        #: delay (False).
        self.send_after_confirmed: bool = True
        # Session socket that means that messages flow to us
        self._flow_in_socket: PipeSocket = None
        self.on_produce_data = self.handle_produce_data
        self.on_accept_data = self.handle_accept_data
        self._msg: FBDPMessage = FBDPMessage()
        self.message_factory = self.__message_factory
        self.handlers.update({MsgType.NOOP: self.handle_noop_msg,
                              MsgType.DATA: self.handle_data_msg,
                              MsgType.CLOSE: self.handle_close_msg,
                              })
    def __message_factory(self, zmsg: TZMQMessage=None) -> Message:
        "Internal message factory"
        self._msg.clear()
        return self._msg
    def _init_new_batch(self, channel: Channel, session: FBDPSession) -> None:
        """Initializes the transmission of a new batch of DATA messages.
        """
        raise NotImplementedError()
    def _send_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Sends next DATA message to the client attached to PIPE_OUTPUT.
        """
        error_code = None
        exc = None
        try:
            self.on_produce_data(channel, session, msg)
            channel.send(msg, session)
            session.transmit -= 1
            if session.transmit > 0:
                if msg.has_ack_req() and self.send_after_confirmed:
                    channel.set_wait_out(False, session)
                elif self.on_get_data.is_set():
                    if not self.on_get_data(channel, session):
                        channel.set_wait_out(False, session)
            else:
                channel.set_wait_out(False, session)
                self._init_new_batch(channel, session)
        except StopError as err:
            error_code = getattr(err, 'code', ErrorCode.ERROR)
            if error_code is not ErrorCode.OK:
                exc = err
        except Exception as err:
            error_code = ErrorCode.INTERNAL_ERROR
            exc = err
        if error_code is not None:
            self.send_close(channel, session, error_code, exc)
    def _on_output_ready(self, channel: Channel) -> None:
        """Event handler called when channel is ready to accept at least one outgoing message
        without blocking (or dropping it).
        """
        for session in list(channel.sessions.values()):
            if session.send_pending:
                msg = self.create_message_for(MsgType.DATA)
                # This is called directly and not via `handle_msg()` and message handler,
                # so it's necessary to handle exceptions like `handle_msg()` does.
                try:
                    self._send_data(channel, session, msg)
                except Exception as exc:
                    try:
                        self.handle_exception(channel, session, msg, exc)
                    except:
                        warnings.warn('Exception raised in exception handler', RuntimeWarning)
    def validate(self, zmsg: TZMQMessage) -> None:
        """Verifies that sequence of ZMQ data frames is a valid protocol message.

        If this validation passes without exception, then `.parse()` of the same message
        must be successful as well.

        Arguments:
            zmsg:   ZeroMQ multipart message.

        Raises:
            InvalidMessageError: If ZMQ message is not a valid protocol message.
        """
        if not zmsg:
            raise InvalidMessageError("Empty message")
        fbdp_header = zmsg[0]
        if len(fbdp_header) != 8:
            raise InvalidMessageError("Message header must be 8 bytes long")
        try:
            fourcc, control_byte, flags, _ = unpack(HEADER_FMT_FULL, fbdp_header)
        except Exception as exp:
            raise InvalidMessageError("Invalid control frame") from exp
        if fourcc != FOURCC:
            raise InvalidMessageError("Invalid FourCC")
        if (control_byte & VERSION_MASK) != self.REVISION:
            raise InvalidMessageError("Invalid protocol version")
        if (flags | 3) > 3:
            raise InvalidMessageError("Invalid flags")
        try:
            message_type = MsgType(control_byte >> 3)
        except ValueError:
            raise InvalidMessageError(f"Illegal message type {control_byte >> 3}")
        if message_type is MsgType.OPEN:
            if len(zmsg) != 2:
                raise InvalidMessageError("OPEN message must have a dataframe")
            try:
                fpb = create_message(PROTO_OPEN)
                fpb.ParseFromString(zmsg[1])
                if not fpb.data_pipe:
                    raise ValueError("Missing 'data_pipe' specification")
                pipe_socket = PipeSocket(fpb.pipe_socket)
                if pipe_socket is PipeSocket.UNKNOWN_PIPE_SOCKET:
                    raise ValueError("Invalid 'pipe_socket'")
                if not fpb.data_format:
                    raise ValueError("Missing 'data_format' specification")
            except Exception as exc:
                raise InvalidMessageError("Invalid data frame for OPEN message") from exc
        elif (message_type is MsgType.CLOSE and len(zmsg) > 1):
            fpb = create_message(PROTO_ERROR)
            for frame in zmsg[1:]:
                fpb.ParseFromString(frame)
                if not fpb.description:
                    raise InvalidMessageError("Missing error description")
        elif (message_type is MsgType.DATA and len(zmsg) > 2):
            raise InvalidMessageError("DATA message may have only one data frame")
        elif (message_type in (MsgType.READY, MsgType.NOOP) and len(zmsg) > 1):
            raise InvalidMessageError("Data frames not allowed for READY and NOOP messages")
    def handle_exception(self, channel: Channel, session: Session, msg: Message, exc: Exception) -> None:
        """Called by `.handle_msg()` on exception in message handler.

        Sends CLOSE message and calls `on_exception` handler.
        """
        error_code = getattr(exc, 'code', ErrorCode.ERROR) if isinstance(exc, StopError) \
            else ErrorCode.INTERNAL_ERROR
        self.send_close(channel, session, error_code, exc)
        super().handle_exception(channel, session, msg, exc)
    def handle_produce_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Default event handler executed when DATA message should be sent to pipe.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: DATA message that will be sent to peer.

        Important:
            The base implementation simply raises StopError with ErrorCode.OK code,
            which closes the pipe normally.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_accept_data(self, channel: Channel, session: FBDPSession, data: bytes) -> None:
        """Default event hander executed when DATA message is received from pipe.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            data: Data received from peer.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Note:
            The ACK-REQUEST in received DATA message is handled automatically by protocol.

        Important:
            The base implementation simply raises StopError with ErrorCode.OK code,
            which closes the pipe normally.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_noop_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process NOOP message received from peer.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        if msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
        self.on_noop(channel, session)
    def handle_data_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process DATA message received from client.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if session.socket is self._flow_in_socket:
            # DATA flow to us (INPUT for server context, OUTPUT for client context)
            if session.transmit is None:
                # Transmission not started, DATA out of band
                raise StopError("Out of band DATA message",
                                code=ErrorCode.PROTOCOL_VIOLATION)
            # ACK before processing?
            if msg.has_ack_req() and not self.confirm_processing:
                # We must create reply message directly to keep received message
                reply = FBDPMessage()
                reply.msg_type = msg.msg_type
                reply.type_data = msg.type_data
                reply.set_flag(MsgFlag.ACK_REPLY)
                if channel.send(msg, session) != 0:
                    raise StopError("ACK-REPLY send failed", code=ErrorCode.ERROR)
            # Process incoming data
            self.on_accept_data(channel, session, msg.data_frame)
            # ACK after processing?
            if msg.has_ack_req() and self.confirm_processing:
                if channel.send(self.create_ack_reply(msg), session) != 0:
                    raise StopError("ACK-REPLY send failed", code=ErrorCode.ERROR)
            session.transmit -= 1
            if session.transmit == 0:
                self._init_new_batch(channel, session)
        else:
            # DATA flow from us (OUTPUT for server context, INPUT for client context)
            if msg.has_ack_reply():
                if (session.transmit > 0) and self.send_after_confirmed:
                    # Re-Initiate transfer to output (via I/O loop) if data are available
                    if not self.on_get_data.is_set() or self.on_get_data(channel, session):
                        channel.set_wait_out(True, session)
                self.on_data_confirmed(channel, session, msg.type_data)
            else:
                # Only client attached to PIPE_INPUT can send DATA messages
                socket: PipeSocket = PipeSocket.OUTPUT \
                    if self._flow_in_socket is PipeSocket.INPUT else PipeSocket.INPUT
                raise StopError(f"DATA message sent to {socket.name} socket",
                                code=ErrorCode.PROTOCOL_VIOLATION)
    def handle_close_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process CLOSE message received from client.

        Calls `on_pipe_closed` and then discards the session.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        try:
            self.on_pipe_closed(channel, session, msg)
        except:
            # We don't want to handle this via `handle_exception` and we're closing
            # the pipe anyway
            pass
        finally:
            channel.discard_session(session)
    def create_message_for(self, msg_type: MsgType, type_data: int=0, flags: MsgFlag=None) -> FBDPMessage:
        """Returns message of particular FBDP message type.

        Arguments:
            msg_type:  Type of message to be created.
            type_data: Message control data.
            flags:     Message flags.
        """
        msg: FBDPMessage = self.message_factory()
        msg.msg_type = msg_type
        msg.type_data = type_data
        if flags is not None:
            msg.flags = flags
        if msg.msg_type is MsgType.OPEN:
            msg.data_frame = create_message(PROTO_OPEN)
        elif msg.msg_type is MsgType.CLOSE:
            msg.data_frame = []
        return msg
    def create_ack_reply(self, msg: FBDPMessage) -> FBDPMessage:
        """Returns new message that is an ACK-REPLY response message.

        Arguments:
            msg: Message to be answered.
        """
        reply = self.create_message_for(msg.msg_type, msg.type_data, msg.flags)
        reply.clear_flag(MsgFlag.ACK_REQ)
        reply.set_flag(MsgFlag.ACK_REPLY)
        return reply
    def send_ready(self, channel: Channel, session: FBDPSession, batch_size: int) -> None:
        """Sends `READY` message.

        Arguments:
            channel: Channel associate with data pipe.
            session: Session associated with transmission.
            batch_size: Requested data transmission batch size.

        Raises:
            StopError: When sending message fails.
        """
        msg = self.create_message_for(MsgType.READY, batch_size)
        if channel.send(msg, session) != 0:
            raise StopError("Broken pipe, can't send READY message", code=ErrorCode.ERROR)
    def send_close(self, channel: Channel, session: FBDPSession, error_code: ErrorCode,
                   exc: Exception=None) -> None:
        """Sends `CLOSE` message, calls `on_pipe_closed` and then discards the session.

        Arguments:
            channel: Channel associate with data pipe.
            session: Session associated with transmission.
            error_code: Error code.
            exc: Exception that caused the error.
        """
        msg = self.create_message_for(MsgType.CLOSE, error_code)
        if exc:
            msg.note_exception(exc)
        try:
            channel.send(msg, session)
            self.on_pipe_closed(channel, session, msg, exc)
        finally:
            channel.discard_session(session)
    @eventsocket
    def on_pipe_closed(self, channel: Channel, session: FBDPSession, msg: FBDPMessage,
                       exc: Exception=None) -> None:
        """Called when CLOSE message is received or sent, to release any resources
        associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.
        """
    @eventsocket
    def on_noop(self, channel: Channel, session: FBDPSession) -> None:
        """Called when NOOP message is received, and after ACK-REPLY (if requested) is send.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
        """
    @eventsocket
    def on_accept_data(self, channel: Channel, session: FBDPSession, data: bytes) -> None:
        """Event executed for CONSUMER to process data received in DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            data: Data received from client.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Note:
            The ACK-REQUEST in received DATA message is handled automatically by protocol.
        """
    @eventsocket
    def on_produce_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Event executed for PRODUCER to store data into outgoing DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            msg: DATA message that will be sent to client.

        The event handler must store the data in `msg.data_frame` attribute. It may also
        set ACK-REQUEST flag and `type_data` attribute.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Note:
            To indicate end of data, raise StopError with ErrorCode.OK code.
        """
    @eventsocket
    def on_data_confirmed(self, channel: Channel, session: FBDPSession, type_data: int) -> None:
        """Event executed for PRODUCER when ACK_REPLY on sent DATA is received.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            type_data: Content of `type_data` field from received DATA message confirmation.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.
        """
    @eventsocket
    def on_get_data(self, channel: Channel, session: FBDPSession) -> bool:
        """Event executed for PRODUCER to query the data source for data availability,
        and for CONSUMER to query whether data could be accepted.

        Important:
            If this event does not have handler assigned, the data source is considered as
            "stable" source that can always produce/consume data (for example data files
            are stable sources).

            For PRODUCERS, handler MUST return True if there are data available for sending.
            If handler returns False, the transmission will be suspended until its resumed
            via `Channel.set_wait_out(True)`.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.
        """
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <class_name>"
        return getattr(self, '_logging_id_', self.__class__.__name__)

class FBDPServer(_FBDP):
    """9/FBDP - Firebird Butler Data Pipe Protocol - Server side.
    """
    def __init__(self, *, session_type: Type[FBDPSession] = FBDPSession):
        """
        Arguments:
            session_type: Class for session objects.
        """
        super().__init__(session_type=session_type)
        # Session socket that means that messages flow to us (client connected to our INPUT)
        self._flow_in_socket: PipeSocket = PipeSocket.INPUT
        #
        self.on_accept_client = self.handle_accept_client
        self.on_get_ready = self.handle_get_ready
        self.on_schedule_ready = self.handle_schedule_ready
        #
        self.handlers.update({MsgType.OPEN: self.handle_open_msg,
                              MsgType.READY: self.handle_ready_msg,
                              })
    def _init_new_batch(self, channel: Channel, session: FBDPSession) -> None:
        """Initializes the transmission of a new batch of DATA messages.

        As we're server, we also have to send READY to the client.
        """
        session.transmit = None
        if (batch_size := self.on_get_ready(channel, session)) == 0:
            self.on_schedule_ready(channel, session)
        if batch_size is not None:
            ready = max(0, self.batch_size if batch_size == -1 else batch_size)
            self.send_ready(channel, session, ready)
            session.await_ready = True
    def handle_accept_client(self, channel: Channel, session: FBDPSession) -> None:
        """Default event handler that raises `StopError` exception with ErrorCode.INTERNAL_ERROR.
        """
        raise StopError("Accept handler not defined", code=ErrorCode.INTERNAL_ERROR)
    def handle_get_ready(self, channel: Channel, session: FBDPSession) -> int:
        """Default event handler that returns -1, unless `on_get_data` event handler is
        assigned and it returns False - then it returns 0.
        """
        if self.on_get_data.is_set() and not self.on_get_data(channel, session):
            return 0
        return -1
    def handle_schedule_ready(self, channel: Channel, session: FBDPSession) -> None:
        """Default event handler that raises `StopError` exception with ErrorCode.INTERNAL_ERROR.

        Note:
            This handler must be reasigned or overriden only when `on_get_ready` event
            handler may return zero.
        """
        raise StopError("READY scheduler not defined", code=ErrorCode.INTERNAL_ERROR)
    def resend_ready(self, channel: Channel, session: FBDPSession) -> None:
        """Send another ready to the client.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
        """
        # It's possible that session could be canceled before resend is called,
        # or transmission was already started, or READY was already sent and awaits
        # response from client. In such case we'll ignore this resend request.
        if (session.routing_id in channel.sessions
            and session.transmit is None
            and not session.await_ready):
            # This is called directly and not via `handle_msg()` and message handler, so
            # it's necessary to handle exceptions like `handle_msg()` does.
            try:
                self._init_new_batch(channel, session)
            except Exception as exc:
                try:
                    self.handle_exception(channel, session, FBDPMessage(), exc)
                except:
                    warnings.warn('Exception raised in exception handler', RuntimeWarning)
        else:
            if session.routing_id not in channel.sessions:
                warnings.warn('resend_ready: senssion cancelled', RuntimeWarning)
            elif session.transmit is not None:
                warnings.warn('resend_ready: transmission already started', RuntimeWarning)
            elif session.await_ready:
                warnings.warn('resend_ready: READY was already sent', RuntimeWarning)
            else:
                warnings.warn('resend_ready: programming error', RuntimeWarning)
    def handle_open_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process OPEN message received from client.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if session.pipe is not None:
            # Client already attached to data pipe, OPEN out of band
            raise StopError("Out of band OPEN message", code=ErrorCode.PROTOCOL_VIOLATION)
        socket = PipeSocket(msg.data_frame.pipe_socket)
        session.pipe = msg.data_frame.data_pipe
        session.socket = socket
        session.data_format = msg.data_frame.data_format
        session.params.update(struct2dict(msg.data_frame.parameters))
        self.on_accept_client(channel, session)
        self._init_new_batch(channel, session)
        channel.on_output_ready = self._on_output_ready
    def handle_ready_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process READY message received from client.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if not session.await_ready:
            # Transmission in progress, READY is out of band
            raise StopError("Out of band READY message",
                            code=ErrorCode.PROTOCOL_VIOLATION)
        session.await_ready = False
        if msg.type_data == 0:
            # Client either confirmed our zero, or is not ready yet.
            self.on_schedule_ready(channel, session)
        else:
            # All green to transmit DATA
            session.transmit = msg.type_data
            if session.socket is PipeSocket.OUTPUT:
                # Initiate transfer to output (via I/O loop)
                channel.set_wait_out(True, session)
    @eventsocket
    def on_accept_client(self, channel: Channel, session: FBDPSession) -> None:
        """Event executed when client connects to the data pipe via OPEN message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The session attributes `data_pipe`, `pipe_socket`, `data_format` and `params`
        contain information sent by client, and the event handler must validate the request.

        If request should be rejected, it must raise the `StopError` exception with `code`
        attribute containing the `ErrorCode` to be returned in CLOSE message.
        """
    @eventsocket
    def on_get_ready(self, channel: Channel, session: FBDPSession) -> int:
        """Event executed to obtain the transmission batch size for the client.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        Returns:
           Number of messages that could be transmitted (batch size):
           * 0 = Not ready to transmit yet
           * n = Ready to transmit 1..<n> messages.
           * -1 = Ready to transmit 1..<protocol batch size> messages.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.
        """
    @eventsocket
    def on_schedule_ready(self, channel: Channel, session: FBDPSession) -> None:
        """The event is executed in order to send the READY message to the client later.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.
        """

class FBDPClient(_FBDP):
    """9/FBDP - Firebird Butler Data Pipe Protocol - Client side.
    """
    def __init__(self, *, session_type: Type[FBDPSession] = FBDPSession):
        """
        Arguments:
            session_type: Class for session objects.
            batch_size: Default batch size.
        """
        super().__init__(session_type=session_type)
        # Session socket that means that messages flow to us (we are connected to server OUTPUT)
        self._flow_in_socket: PipeSocket = PipeSocket.OUTPUT
        self.on_server_ready = self.handle_server_ready
        #
        self.handlers.update({MsgType.OPEN: self.handle_open_msg,
                              MsgType.READY: self.handle_ready_msg,
                              })
    def handle_server_ready(self, channel: Channel, session: FBDPSession, batch_size: int) -> int:
        """Default event handler that returns -1, unless `on_get_data` event handler is
        assigned and it returns False - then it returns 0.
        """
        if self.on_get_data.is_set() and not self.on_get_data(channel, session):
            return 0
        return -1
    def _init_new_batch(self, channel: Channel, session: FBDPSession) -> None:
        """Initializes the transmission of a new batch of DATA messages.
        """
        session.transmit = None
    def accept_new_session(self, channel: Channel, routing_id: RoutingID,
                           msg: FBDPMessage) -> bool:
        """Validates incoming message that initiated new session/transmission.

        Arguments:
            channel:    Channel that received the message.
            routing_id: Routing ID of the sender.
            msg:        Received message.

        Returns:
            Always False (transmission must be initiated by Client).
        """
        return False
    def connect_with_session(self, channel: Channel) -> bool:
        """Called by :meth:`Channel.connect` to determine whether new session should be
        associated with connected peer.

        As FBDP require that connecting peers must send OPEN message to initiate
        transmission, it always returns True.
        """
        return True
    def handle_open_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """OPEN message received from server is violation of the protocol.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        raise StopError("OPEN message received from server", ErrorCode.PROTOCOL_VIOLATION)
    def handle_ready_msg(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Process READY message received from server.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if session.transmit is not None:
            # Transmission in progress, READY is out of band
            raise StopError("Out of band READY message",
                            code=ErrorCode.PROTOCOL_VIOLATION)
        if msg.type_data > 0:
            # Server is ready
            batch_size = self.on_server_ready(channel, session, msg.type_data)
            result = max(0, min(msg.type_data, self.batch_size if batch_size == -1 else batch_size))
            self.send_ready(channel, session, result)
            if result > 0:
                # We are ready to transmit as well
                session.transmit = result
                if session.socket is PipeSocket.INPUT:
                    # Initiate transfer to server (via I/O loop)
                    channel.set_wait_out(True, session)
        else:
            # Server is not ready, but we must send READY(0) back to confirm we've got it!
            self.send_ready(channel, session, 0)
    def send_open(self, channel: Channel, session: FBDPSession, data_pipe: str,
                  pipe_socket: PipeSocket, data_format: str, parameters: Dict=None) -> None:
        """Sends `OPEN` message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with transmission.
            data_pipe: Data pipe identification.
            pipe_socket: Connected pipe socket.
            data_format: Required data format.
            parameters: Data pipe parameters.

        Raises:
            StopError: When sending message fails.
        """
        msg = self.create_message_for(MsgType.OPEN)
        msg.data_frame.data_pipe = data_pipe
        msg.data_frame.pipe_socket = pipe_socket.value
        msg.data_frame.data_format = data_format
        if parameters:
            msg.data_frame.parameters.CopyFrom(dict2struct(parameters))
        if channel.send(msg, session) != 0:
            raise StopError("Broken pipe, can't send OPEN message", code=ErrorCode.ERROR)
        channel.on_output_ready = self._on_output_ready
        session.pipe = data_pipe
        session.socket = pipe_socket
        session.data_format = data_format
        if parameters:
            session.params.update(parameters)
        self.on_init_session(channel, session)
    @eventsocket
    def on_server_ready(self, channel: Channel, session: FBDPSession, batch_size: int) -> int:
        """Event executed to negotiate the transmission batch size with server.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with server.
            batch_size: Max. batch size accepted by server.

        Returns:
           Number of messages that could be transmitted (batch size):
           * 0 = Not ready to transmit yet
           * n = Ready to transmit 1..<n> messages.
           * -1 = Ready to transmit 1..<protocol batch size> messages.

        Important:
            The returned value will be used ONLY when it's smaller than `batch_size`.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.
        """
    @eventsocket
    def on_init_session(self, channel: Channel, session: FBDPSession) -> None:
        """Event executed from `send_open()` to set additional information to newly
        created session instance.
        """



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/init.py
# DESCRIPTION:    Script for Saturnin platform deployment initialization
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin platform deployment initialization script


"""

from __future__ import annotations
from traceback import format_exception_only
from argparse import ArgumentParser, Namespace # , ArgumentDefaultsHelpFormatter
from pathlib import Path
#from site import getsitepackages
#from saturnin.core import Error
from saturnin.base.site import site, CONFIG_HDR

PROG_NAME = 'saturnin-init'



class Script:
    """Saturnin platform initialization.
    """
    def __init__(self):
        ""
    def out(self, msg: str) -> None:
        ""
        print(msg, end='', flush=True)
    def init_directories(self, args: Namespace) -> None:
        ""
        def ensure_dir(name: str, path: Path):
            self.out(f"{name}: {path} ... ")
            if not path.exists():
                path.mkdir(parents=True)
            self.out("OK\n")

        self.out('Saturnin directories...\n')
        if site.scheme.has_home_env():
            print(f"  SATURNIN_HOME is set to     : {site.scheme.home}")
        else:
            print("  SATURNIN_HOME env. variable not defined")
        ensure_dir("  Saturnin configuration      ", site.scheme.config)
        ensure_dir("  Saturnin data               ", site.scheme.data)
        ensure_dir("  Run-time data               ", site.scheme.run_data)
        ensure_dir("  Log files                   ", site.scheme.logs)
        ensure_dir("  Temporary files             ", site.scheme.tmp)
        ensure_dir("  Cache                       ", site.scheme.cache)
        ensure_dir("  User-specific configuration ", site.scheme.user_config)
        ensure_dir("  User-specific data          ", site.scheme.user_data)
        #ensure_dir("  Python virtual environment  ", site.scheme.venv)
        ensure_dir("  PID files                   ", site.scheme.pids)
    def create_config_files(self, args: Namespace) -> None:
        ""
        self.out("Creating configuration files...\n")
        self.out(f"  Writing : {site.scheme.site_conf} ... ")
        cfg = CONFIG_HDR + site.config.get_config()
        site.scheme.site_conf.write_text(cfg)
        self.out("OK\n")
        self.out(f"  Writing : {site.scheme.user_conf} ... ")
        site.scheme.user_conf.write_text(cfg)
        self.out("OK\n")
    def run(self) -> None:
        ""
        parser: ArgumentParser = ArgumentParser(PROG_NAME, description=self.__doc__)
        parser.add_argument('--make-dirs', action='store_true',
                            help="Create saturnin directories", default=None)
        parser.add_argument('--config-files', action='store_true',
                            help="Create configuration files", default=None)
        try:
            args = parser.parse_args()
            no_args = True
            for k in vars(args):
                if getattr(args, k) is not None:
                    no_args = False
                    break
            if no_args or args.make_dirs:
                self.init_directories(args)
            if no_args or args.config_files:
                self.create_config_files(args)
        except Exception as exc:
            self.out("ERROR\n")
            msg = ''.join(format_exception_only(type(exc), exc))
            _, msg = msg.split(': ', 1)
            parser.exit(1, msg)

def main():
    script = Script()
    script.run()

if __name__ == '__main__':
    main()




#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/pkg.py
# DESCRIPTION:    Saturnin package manager commands
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin package manager commands


"""

from __future__ import annotations
from typing import Dict, List, Optional
import sys
import toml
import zipfile
import subprocess
from pathlib import Path
from packaging import version
from argparse import ArgumentParser, Namespace, REMAINDER
from configparser import ConfigParser, ExtendedInterpolation
from saturnin.base import Error, site
from saturnin.lib.command import Command, CommandManager
from saturnin.lib.pyproject import PyProjectTOML

PKG_TOML = 'pyproject.toml'
SETUP_CFG = 'setup.cfg'
SEC_OPTIONS = 'options'
SEC_ENTRYPOINTS = 'options.entry_points'

# Component registry TOML structure
#
# [[components]]
# uid = component GUID
# component-type = 'service' or 'app'
# package = normalized component name
# name = component name
# version = component version
# description = component description
# descriptor = locator string for component descriptor (incl. top level package)
# top-level = top level package with component

CMP_UID = 'uid'
CMP_PACKAGE = 'package'
CMP_NAME = 'name'
CMP_VERSION = 'version'
CMP_DESCRIPTION = 'description'
CMP_DESCRIPTOR = 'descriptor'
CMP_TOPLEVEL = 'top-level'

class BasePkgCommand(Command):
    """Base class for saturnin-pkg commands.
    """
    def __init__(self, name: str, description: str):
        super().__init__(name, description)
        self.components: List[Dict]= []
    def load_components(self) -> None:
        if site.scheme.site_components_toml.is_file():
            self.components.extend(toml.load(site.scheme.site_components_toml)['components'])
    def save_components(self) -> None:
        site.scheme.site_components_toml.write_text(toml.dumps({'components': self.components}))
    def get_new_toplevel(self) -> str:
        ids = [int(s[1:]) for s in (c['top-level'] for c in self.components)]
        i = 1
        while i in ids:
            i += 1
        return f'C{i:05}'
    def find_component_index(self, *, uid: str=None, name: str=None, package: str=None) -> int:
        if uid is not None:
            key = CMP_UID
            value = uid
        elif name is not None:
            key = CMP_NAME
            value = name
        elif package is not None:
            key = CMP_PACKAGE
            value = package
        else:
            raise ValueError("Missing lookup key specification")
        for i in range(len(self.components)):
            if value == self.components[i][key]:
                return i
        return -1
    def find_component(self, *, uid: str=None, name: str=None, package: str=None) -> Optional[Dict]:
        i = self.find_component_index(uid=uid, name=name, package=package)
        return None if i < 0 else self.components[i]

class InstallCommand(BasePkgCommand):
    """saturnin-pkg INSTALL command.

    Installation of Saturnin packages.
    """
    def __init__(self):
        super().__init__('install', "Install Saturnin ZIP package.")
    def copy_tree(self, package: Path, path: zipfile.Path, skip_list: List[str]=[]):
        ""
        package.mkdir(exist_ok=True)
        for item in path.iterdir():
            if item.is_dir():
                self.copy_tree(package / item.name, item)
            elif item.name not in skip_list:
                target = package / item.name
                target.write_bytes(item.read_bytes())
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
        parser.add_argument('package', help="Saturnin ZIP package")
        parser.add_argument('args', nargs=REMAINDER, help="Arguments for 'pip install'")
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        pkg = Path(args.package)
        if not pkg.is_file():
            raise ValueError("File not found")
        self.load_components()
        with zipfile.ZipFile(pkg) as zf:
            root = zipfile.Path(zf)
            pkg_info = root / PKG_TOML
            if pkg_info.exists():
                # Single-service package
                self.install_package(args.args, zf, root)
            else:
                # Service bundle, each subdirectory should contain one service
                for pkg in root.iterdir():
                    pkg_info = pkg / PKG_TOML
                    if pkg_info.exists():
                        self.install_package(args.args, zf, pkg)
    def remove_dir(self, path: Path) -> None:
        if isinstance(path, str):
            path = Path(path)
        paths = list(path.rglob('*'))
        paths.reverse()
        for p in paths:
            if p.is_file():
                p.unlink()
            else:
                p.rmdir()
        path.rmdir()
    def install_package(self, args: List, zfile: zipfile.ZipFile, path: zipfile.Path):
        work_dir = site.scheme.tmp / 'import'
        if work_dir.exists():
            self.remove_dir(work_dir)
        work_dir.mkdir()
        #
        root_files = [PKG_TOML]
        # pyproject.toml
        pkg_file: zipfile.Path = path / PKG_TOML
        if not pkg_file.exists():
            raise Error(f"Invalid package: File {_file} not found")
        toml_data = pkg_file.read_text()
        pkg_toml: PyProjectTOML = PyProjectTOML(toml.loads(toml_data))
        self.out(f"Found: {pkg_toml.original_name}-{pkg_toml.version}\n")
        #
        add_component = False
        cmp: Dict = self.find_component(uid=pkg_toml.uid)
        if cmp is None:
            cmp = {}
            top_level = self.get_new_toplevel()
            cmp[CMP_UID] = pkg_toml.uid
            cmp[CMP_NAME] = pkg_toml.original_name
            cmp[CMP_PACKAGE] = pkg_toml.name
            cmp[CMP_VERSION] = pkg_toml.version
            cmp[CMP_DESCRIPTION] = pkg_toml.description
            cmp[CMP_DESCRIPTOR] = f'{top_level}.{pkg_toml.descriptor}'
            cmp[CMP_TOPLEVEL] = top_level
            add_component = True
        else:
            top_level = cmp[CMP_TOPLEVEL]
            vn = version.parse(pkg_toml.version)
            vo = version.parse(cmp[CMP_VERSION])
            if vn < vo:
                self.out(f"Newer version {cmp[CMP_VERSION]} already installed\n")
                return
            elif vn == vo:
                self.out(f"Version {cmp[CMP_VERSION]} already installed\n")
                return
            cmp[CMP_NAME] = pkg_toml.original_name
            cmp[CMP_PACKAGE] = pkg_toml.name
            cmp[CMP_VERSION] = pkg_toml.version
            cmp[CMP_DESCRIPTION] = pkg_toml.description
            cmp[CMP_DESCRIPTOR] = f'{top_level}.{pkg_toml.descriptor}'
        # write toml
        target: Path = work_dir / PKG_TOML
        target.write_text(toml_data)
        # setup.cfg
        cfg: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
        pkg_file = path / SETUP_CFG
        if pkg_file.exists():
            root_files.append(SETUP_CFG)
            cfg.read_string(pkg_file.read_text())
        pkg_toml.make_setup_cfg(cfg)
        if not cfg.has_section(SEC_OPTIONS):
            cfg.add_section(SEC_OPTIONS)
        cfg[SEC_OPTIONS]['packages'] = top_level
        if not cfg.has_section(SEC_ENTRYPOINTS):
            cfg.add_section(SEC_ENTRYPOINTS)
        if pkg_toml.component_type == 'service':
            cfg[SEC_ENTRYPOINTS]['saturnin.service'] = f'\n{cmp[CMP_UID]} = {cmp[CMP_DESCRIPTOR]}'
        elif pkg_toml.component_type == 'application':
            cfg[SEC_ENTRYPOINTS]['saturnin.application'] = f'\n{cmp[CMP_UID]} = {cmp[CMP_DESCRIPTOR]}'
        target = work_dir / SETUP_CFG
        with target.open('w') as f:
            cfg.write(f)
        # README
        if pkg_toml.readme is not None:
            readme_file = None
            if isinstance(pkg_toml.readme, str):
                readme_file = pkg_toml.readme
            elif 'file' in pkg_toml.readme:
                readme_file = pkg_toml.readme['file']
            if readme_file is not None:
                root_files.append(readme_file)
                target = work_dir / readme_file
                pkg_file = path / readme_file
                target.write_bytes(pkg_file.read_bytes())
        # LICENSE
        if pkg_toml.license is not None:
            if (license_file := pkg_toml.license.get('file')) is not None:
                root_files.append(license_file)
                target = work_dir / license_file
                pkg_file = path / license_file
                target.write_bytes(pkg_file.read_bytes())
        # copy files to top-level
        self.copy_tree(work_dir / top_level, path, root_files)
        # Run pip to install prepared package
        pip_cmd = [str(site.pip_path), 'install']
        pip_cmd.extend(args)
        pip_cmd.append(str(work_dir))
        try:
            subprocess.run(pip_cmd, stdout=sys.stdout, stderr=sys.stderr,
                           check=True, text=True)
        except subprocess.CalledProcessError:
            print('Component installation failed')
            raise
        else:
            if add_component:
                self.components.append(cmp)
            self.save_components()

class UninstallCommand(BasePkgCommand):
    """saturnin-pkg UNINSTALL command.

    Uninstallation of Saturnin components.
    """
    def __init__(self):
        super().__init__('uninstall', "Uninstall Saturnin component.")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
        parser.add_argument('component', help="Component name")
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        self.load_components()
        i = self.find_component_index(name=args.component)
        if i < 0:
            raise ValueError("Component not found.")
        cmp = self.components.pop(i)
        # Run pip to uninstall package
        try:
            subprocess.run([str(site.pip_path), 'uninstall', '-y', cmp[CMP_PACKAGE]],
                           stdout=sys.stdout, stderr=sys.stderr,
                           check=True, text=True)
        except subprocess.CalledProcessError:
            print('Component uninstallation failed')
            raise
        #
        self.save_components()

class ListCommand(BasePkgCommand):
    """saturnin-pkg LIST command.

    List installed Saturnin components.
    """
    def __init__(self):
        super().__init__('list', "List installed Saturnin components.")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        self.load_components()
        self.print_table(['Component', 'Version', 'Package'],
                         [[cmp[CMP_NAME], cmp[CMP_VERSION], cmp[CMP_PACKAGE]]
                          for cmp in self.components])

class PipCommand(BasePkgCommand):
    """saturnin-pkg PIP command.

    Run pip from Saturnin virtual environment.
    """
    def __init__(self):
        super().__init__('pip', "Run pip from Saturnin virtual environment.")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
        parser.add_argument('args', nargs=REMAINDER, help="Arguments for pip")
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        # Run pip to uninstall package
        pip_cmd = [str(site.pip_path)]
        pip_cmd.extend(args.args)
        try:
            subprocess.run(pip_cmd, stdout=sys.stdout, stderr=sys.stderr,
                           check=True, text=True)
        except subprocess.CalledProcessError:
            print('pip execution failed')
            raise



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/controler.py
# DESCRIPTION:    Service controllers
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin service controllers
"""

from __future__ import annotations
from typing import Dict, List, cast
import sys
import uuid
import zmq
import signal
import warnings
import ctypes
import weakref
from time import monotonic
from threading import Thread
from configparser import ConfigParser
from firebird.base.config import UUIDOption
from firebird.base.trace import TracedMixin
from saturnin.base import ZMQAddress, load, Error, ServiceError, PeerDescriptor, Outcome, \
     ServiceDescriptor, Direction, ChannelManager, PairChannel, INVALID, Component, Config
from saturnin.protocol.iccp import ICCPComponent, ICCPController, ICCPMessage, MsgType

#: Service control channel name
SVC_CTRL = sys.intern('iccp')

class ServiceExecConfig(Config):
    """Service executor configuration.
    """
    def __init__(self, name: str):
        super().__init__(name)
        self.agent: UUIDOption = UUIDOption('agent', "Agent UID", required=True)

class ServiceController(TracedMixin):
    """
    """
    def __init__(self, service: ServiceDescriptor, *, name: str=None,
                 peer_uid: uuid.UUID=None, manager: ChannelManager=None):
        """
        Arguments:
            service: Service to start.
            name:  Container name.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
            manager: ChannelManager to be used.
        """
        self.outcome: Outcome = Outcome.UNKNOWN
        self.details: List[str] = []
        self.log_context = None
        self.name: str = service.agent.name if name is None else name
        self.peer_uid: uuid.UUID = peer_uid
        self.service: ServiceDescriptor = service
        self.peer: PeerDescriptor = None
        self.endpoints: Dict[str, List[ZMQAddress]] = {}
        self.config: Config = service.config()
        self.ctrl_addr: ZMQAddress = ZMQAddress(f'inproc://{uuid.uuid1().hex}')
        self.mngr: ChannelManager = manager
        self._ext_mngr: bool = manager is not None
    def __str__(self):
        return self.logging_id
    def configure(self, config: ConfigParser, section: str=None) -> None:
        """Loads and validates service configuration.

        Arguments:
            config: ConfigParser instance with service configuration.
            section: Name of section with service configuration. If not present, uses
                     `ServiceContainer.name` value.
        """
        self.config.load_config(config, self.name if section is None else section)
        self.config.validate()
    @property
    def logging_id(self) -> str:
        "Returns qualified class name and agent name."
        return f'{self.__class__.__qualname__}[{self.service.agent.name}]'

class DirectController(ServiceController):
    """Service controller that starts the service in current thread.

    Although ICCP is used to start/stop the service, it's not possible to perform any ICCP
    interactions with the service while it's running.

    Important:
        The service could be stopped only via automatically installed SIGINT handler.
    """
    def stop_signal_handler(self, signum: int, param) -> None:
        """The `signal.signal()` SIGINT handler that sends ICCP STOP message to the service.
        """
        chn: PairChannel = self.mngr.channels[SVC_CTRL]
        chn.send(cast(ICCPController, chn.protocol).stop_msg(), chn.session)
    def handle_stop_controller(self, exc: Exception) -> None:
        """Called when controller should stop its operation due to error condition.

        Arguments:
           exc: Exception that describes the reason why component should stop.
        """
        raise ServiceError("Internal controller error") from exc
    def start(self, *, timeout: int=10000) -> None:
        """Start the service.

        Arguments:
            timeout: Only for compatibility with ThreadController. The value is ignored.

        Important:
            Will not return until service is stopped via SIGINT, or exception is raised.

        Raises:
            ServiceError: On error in communication with service.
        """
        if not self._ext_mngr:
            self.mngr = ChannelManager(zmq.Context.instance())
            self.mngr.log_context = weakref.proxy(self)
        iccp = ICCPController()
        iccp.on_stop_controller = self.handle_stop_controller
        chn: PairChannel = self.mngr.create_channel(PairChannel, SVC_CTRL, iccp,
                                                    wait_for=Direction.IN,
                                                    sock_opts={'rcvhwm': 5, 'sndhwm': 5,})
        chn.protocol.log_context = self.log_context
        #
        factory = load(self.service.factory)
        svc: Component = factory(zmq.Context.instance(), self.service)
        svc.initialize(self.config)
        # 1st phase successful
        self.mngr.warm_up()
        chn.bind(self.ctrl_addr)
        try:
            # 2nd phase
            svc.warm_up(self.ctrl_addr) # Creates sockets, connects to `iccp`
            result = chn.wait(1000)
            if result == Direction.IN:
                msg: ICCPMessage = chn.receive()
                if msg is INVALID:
                    raise ServiceError("Invalid response from service")
                elif msg.msg_type is MsgType.READY:
                    self.peer = msg.peer.copy()
                    self.endpoints = msg.endpoints.copy()
                elif msg.msg_type is MsgType.ERROR:
                    raise ServiceError(msg.error)
                else:
                    raise ServiceError("ICCP protocol error - unexpected message")
            else:
                raise ServiceError("Service did not started in time")
            # All green to run the service
            old = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, self.stop_signal_handler)
            try:
                svc.run()
                result = chn.wait(100)
                if result == Direction.IN:
                    msg: ICCPMessage = chn.receive()
                    if msg is INVALID:
                        raise ServiceError("Invalid response from service")
                    elif msg.msg_type is MsgType.ERROR:
                        raise ServiceError(msg.error)
                else:
                    warnings.warn("Service shutdown not confirmed", RuntimeWarning)
            finally:
                signal.signal(signal.SIGINT, old)
        finally:
            if not self._ext_mngr:
                self.mngr.shutdown(forced=True)

def service_thread(service: ServiceDescriptor, config: Config, ctrl_addr: ZMQAddress,
                   peer_uid: uuid.UUID=None):
    """Thread target code to run the service.

    Arguments:
        svc_descriptor: Service descriptor.
        config:         Service configuration.
        ctrl_addr:      Address for control ZMQ socket.
        peer_uid:       Peer ID.
    """
    try:
        ctx = zmq.Context.instance()
        pipe = ctx.socket(zmq.DEALER)
        pipe.CONNECT_TIMEOUT = 5000 # 5sec
        pipe.IMMEDIATE = 1
        pipe.LINGER = 5000 # 5sec
        pipe.SNDTIMEO = 5000 # 5sec
        try:
            factory = load(service.factory)
            svc: Component = factory(zmq.Context.instance(), service, peer_uid=peer_uid)
            svc.initialize(config)
        except Exception as exc:
            pipe.connect(ctrl_addr)
            iccp = ICCPComponent()
            pipe.send_multipart(iccp.error_msg(exc).as_zmsg())
            raise
        finally:
            pipe.close(100)
        svc.warm_up(ctrl_addr) # Creates sockets, connects to `iccp`
        svc.run()
    except:
        pass

class ThreadController(ServiceController):
    """Service controller that starts the service in separate thread.
    """
    def __init__(self, service: ServiceDescriptor, *, name: str=None,
                 peer_uid: uuid.UUID=None, manager: ChannelManager=None):
        """
        Arguments:
            service: Service to start.
            name:  Container name.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
        """
        super().__init__(service, name=name, peer_uid=peer_uid, manager=manager)
        self.runtime: Thread = None
    def handle_stop_controller(self, exc: Exception) -> None:
        """Called when controller should stop its operation due to error condition.

        Arguments:
           exc: Exception that describes the reason why component should stop.
        """
        raise ServiceError("Internal controller error") from exc
    def is_running(self) -> bool:
        """Returns True if service is running.
        """
        if self.runtime is None:
            return False
        if self.runtime.is_alive():
            return True
        # It's dead, so dispose the runtime
        self.runtime = None
        return False
    def start(self, *, timeout: int=10000) -> None:
        """Start the service.

        Arguments:
            timeout: Timeout (in milliseconds) to wait for service to report it's ready.

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When timeout expires.
        """
        if not self._ext_mngr:
            self.mngr = ChannelManager(zmq.Context.instance())
            self.mngr.log_context = weakref.proxy(self)
        iccp = ICCPController()
        iccp.on_stop_controller = self.handle_stop_controller
        chn: PairChannel = self.mngr.create_channel(PairChannel, f'{self.name}.{SVC_CTRL}',
                                                    iccp, wait_for=Direction.IN,
                                                    sock_opts={'rcvhwm': 5, 'sndhwm': 5,})
        chn.protocol.log_context = self.log_context
        self.mngr.warm_up()
        chn.bind(self.ctrl_addr)
        #
        self.runtime = Thread(target=service_thread, name=self.name,
                              args=(self.service, self.config, self.ctrl_addr, self.peer_uid),
                              daemon=False)
        self.runtime.start()
        #
        try:
            result = chn.wait(timeout)
            if result == Direction.IN:
                msg: ICCPMessage = chn.receive()
                if msg is INVALID:
                    raise ServiceError("Invalid response from service")
                elif msg.msg_type is MsgType.READY:
                    self.peer = msg.peer.copy()
                    self.endpoints = msg.endpoints.copy()
                elif msg.msg_type is MsgType.ERROR:
                    raise ServiceError(msg.error)
                else:
                    raise ServiceError("ICCP protocol error - unexpected message")
            elif not self.is_running():
                raise ServiceError("Service start failed for unknown reason")
            else:
                raise TimeoutError("Service did not started in time")
        except Exception:
            if not self._ext_mngr:
                self.mngr.shutdown(forced=True)
            raise
    def stop(self, *, timeout: int=10000) -> None:
        """Stop the service. Does nothing if service is not running.

        Arguments:
            timeout: None (infinity), or timeout (in milliseconds) for the operation.

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When service does not stop in time.
        """
        s = monotonic()
        try:
            chn: PairChannel = self.mngr.channels[f'{self.name}.{SVC_CTRL}']
            if self.is_running():
                chn.send(cast(ICCPController, chn.protocol).stop_msg(), chn.session)
                result = chn.wait(timeout)
                if result == Direction.IN:
                    msg: ICCPMessage = chn.receive()
                    if msg is INVALID:
                        self.outcome = Outcome.ERROR
                        self.details = ["Invalid response from service"]
                        raise ServiceError("Invalid response from service")
                    elif msg.msg_type is MsgType.ERROR:
                        self.outcome = Outcome.ERROR
                        self.details = [msg.error]
                        raise ServiceError(msg.error)
                    elif msg.msg_type is MsgType.FINISHED:
                        self.outcome = msg.outcome
                        self.details = msg.details
                else:
                    warnings.warn("Service shutdown not confirmed", RuntimeWarning)
                #
                if self.is_running():
                    e = monotonic()
                    if timeout is not None:
                        timeout = timeout - int((e-s) * 1000)
                        if timeout < 0:
                            timeout = 0
                    self.runtime.join(timeout)
                    if self.runtime.is_alive():
                        raise TimeoutError("The service did not stop in time")
            else:
                result = chn.wait(0)
                if result == Direction.IN:
                    msg: ICCPMessage = chn.receive()
                    if msg is INVALID:
                        self.outcome = Outcome.ERROR
                        self.details = ["Invalid response from service"]
                        raise ServiceError("Invalid response from service")
                    elif msg.msg_type is MsgType.ERROR:
                        self.outcome = Outcome.ERROR
                        self.details = [msg.error]
                        raise ServiceError(msg.error)
                    elif msg.msg_type is MsgType.FINISHED:
                        self.outcome = msg.outcome
                        self.details = msg.details
        finally:
            if not self._ext_mngr:
                self.mngr.shutdown(forced=True)
    def terminate(self) -> None:
        """Terminate the service.

        Terminate should be called ONLY when call to stop() (with sensible timeout) fails.
        Does nothing when service is not running.

        Raises:
            Error:  When service termination fails.
        """
        if self.is_running():
            tid = ctypes.c_long(self.runtime.ident)
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(SystemExit))
            if res == 0:
                raise Error("Service termination failed due to invalid thread ID.")
            if res != 1:
                # if it returns a number greater than one, you're in trouble,
                # and you should call it again with exc=NULL to revert the effect
                ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
                raise Error("Service termination failed due to PyThreadState_SetAsyncExc failure")
    def join(self, timeout=None) -> None:
        """Wait until service stops.

        Arguments:
            timeout: Floating point number specifying a timeout for the operation in
                     seconds (or fractions thereof).
        """
        if self.runtime:
            self.runtime.join(timeout)



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/client.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Service clients
# CREATED:        15.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin base module for implementation of Firebird Butler Service clients


"""

from __future__ import annotations
import os
import platform
from struct import pack
from uuid import UUID
from saturnin.base import Error, ZMQAddress, Channel, TIMEOUT, INVALID, AgentDescriptor, \
     PeerDescriptor
from saturnin.protocol.fbsp import FBSPClient, FBSPSession, FBSPMessage, \
     WelcomeMessage, ErrorMessage, MsgType

class Token():
    """FBSP message token generator.
    """
    def __init__(self, value: int=0):
        self._value: int = value
    def next(self) -> bytes:
        """Returns next message token.
        """
        result = pack('!Q', self._value)
        self._value += 1
        return result

class ServiceClient:
    """Base class for Firebird Butler Service clients.
    """
    def __init__(self, timeout: int=1000):
        """
        Arguments:
            timeout: The timeout (in milliseconds) to wait for message.
        """
        self.token: Token = Token()
        self.timeout: int = timeout
        self.channel: Channel = None
        self.session: FBSPSession = None
        self.protocol: FBSPClient = None
    def open(self, channel: Channel, address: ZMQAddress, agent: AgentDescriptor,
             peer_uid: UUID) -> None:
        """Open connection to Firebird Butler service.

        Arguments:
            channel: Channel used for communication with service.
            address: Service endpoint address.
            agent: Client agent identification.
            peer_uid: Client peer ID.
        """
        assert isinstance(channel.protocol, FBSPClient)
        self.channel = channel
        self.session = channel.connect(address)
        self.protocol = channel.protocol
        self.protocol.send_hello(channel, self.session, agent,
                                 PeerDescriptor(peer_uid, os.getpid(), platform.node()),
                                 self.token.next())
        msg = self.channel.receive(self.timeout)
        if isinstance(msg, ErrorMessage):
            raise self.protocol.exception_for(msg)
        elif msg is TIMEOUT:
            raise TimeoutError()
        elif msg is INVALID:
            raise Error("Invalid response from service")
        elif not isinstance(msg, WelcomeMessage):
            raise Error(f"Unexpected {msg.msg_type.name} message from service")
    def send(self, msg: FBSPMessage) -> None:
        """Send message to the service.

        Arguments:
            msg: Message to be sent.
        """
        self.channel.send(msg, self.session)
    def receive(self) -> FBSPMessage:
        """Receive one message from service.

        Raises:
            TimeoutError: When timeout expires.
            Error: When `Channel.receive()` returns INVALID sentinel, or service closes
                   connection with CLOSE message.
        """
        msg = self.channel.receive(self.timeout)
        if isinstance(msg, ErrorMessage):
            raise self.protocol.exception_for(msg)
        elif msg is TIMEOUT:
            raise TimeoutError()
        elif msg is INVALID:
            raise Error("Invalid response from service")
        elif msg.msg_type is MsgType.CLOSE:
            raise Error("Connection closed by service")
        return msg
    def close(self) -> None:
        """Close connection to service.
        """
        self.protocol.send_close(self.channel, self.session)
        self.channel.discard_session(self.session)
        self.session = None
        self.channel = None
    @property
    def connected(self) -> bool:
        """True if client is connected to service.
        """
        return (self.session is not None) and (self.session.greeting is not None)






#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/site.py
# DESCRIPTION:    Saturnin site
# CREATED:        7.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin site


"""

from __future__ import annotations
from typing import List, Optional
import sys
import os
from pathlib import Path
from configparser import ConfigParser, ExtendedInterpolation
from firebird.base.config import DirectoryScheme, get_directory_scheme, Config, StrOption
from firebird.base.logging import LoggingIdMixin

SATURNIN_CFG = 'saturnin.conf'

CONFIG_HDR = """;
; A configuration file consists of sections, each led by a [section] header, followed by
; key/value entries separated by = or : string. Section names are case sensitive but keys
; are not. Leading and trailing whitespace is removed from keys and values. Values can be
; omitted, in which case the key/value delimiter may also be left out. Values can also
; span multiple lines, as long as they are indented deeper than the first line of the value.
;
; Configuration files may include comments, prefixed by # and ; characters. Comments may
; appear on their own on an otherwise empty line, possibly indented.
;
; Values can contain ${section:option} format strings which refer to other values.
; If the section: part is omitted, interpolation defaults to the current section (and possibly
; the default values from the special DEFAULT section). Interpolation can span multiple levels.

"""

class SaturninScheme(DirectoryScheme):
    """Saturnin platform directory scheme.
    """
    def __init__(self):
        super().__init__('saturnin')
        self.dir_map.update(get_directory_scheme('saturnin').dir_map)
    @property
    def pids(self) -> Path:
        """Path to directory with PID files for running daemons.
        """
        return self.run_data / 'pids'
    @property
    def site_components_toml(self) -> Path:
        """Saturnin package registry file.
        """
        return self.data / 'components.toml'
    @property
    def site_conf(self) -> Path:
        """Saturnin site configuration file.
        """
        return self.config / SATURNIN_CFG
    @property
    def user_conf(self) -> Path:
        """Saturnin user configuration file.
        """
        return self.user_config / SATURNIN_CFG

class SaturninConfig(Config):
    """Saturnin (a Firebird Butler platform) configuration.
    """
    def __init__(self):
        super().__init__('saturnin')
        self.editor: StrOption = StrOption('editor', "External editor",
                                           default=os.getenv('EDITOR'))

class SiteManager(LoggingIdMixin):
    """Saturnin site manager.
    """
    def __init__(self):
        # Path to pip
        self.pip_path: Path = 'pip'
        # Set SATUTNIN_HOME if defined in virtual environment
        if self.is_virtual():
            path: Path = self.venv / '.saturnin-home'
            if path.is_file():
                os.environ['SATURNIN_HOME'] = path.read_text()
            path = self.venv / '.saturnin-bin'
            if path.is_file():
                self.pip_path = Path(path.read_text()) / 'pip'
        #: Saturnin directory scheme
        self.scheme: SaturninScheme = SaturninScheme()
        #: Saturnin configuration
        self.config: SaturninConfig = SaturninConfig()
        #: Used configuration files
        self.used_config_files: List[Path] = []
    def load_configuration(self) -> None:
        """Loads the configuration from configuration files.
        """
        parser: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
        for f in parser.read([self.scheme.site_conf,
                              self.scheme.user_conf]):
            self.used_config_files.append(Path(f))
        if parser.has_section('saturnin'):
            self.config.load_config(parser)
    def initialize(self) -> None:
        """Saturnin site initialization.
        """
        self.load_configuration()
    def is_virtual(self) -> bool:
        """Returns True if site runs in a virtual environtment.
        """
        # Check supports venv && virtualenv >= 20.0.0
        return getattr(sys, 'base_prefix', sys.prefix) != sys.prefix
    @property
    def venv(self) -> Optional[Path]:
        """Path to Saturnin virtual environment.
        """
        return Path(sys.prefix) if self.is_virtual() else None

#: Saturnin site manager
site: SiteManager = SiteManager()
site.initialize()



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/cli.py
# DESCRIPTION:    Script for Saturnin CLI manager
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin CLI manager script


"""

from __future__ import annotations
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from saturnin.lib.command import CommandManager

PROG_NAME = 'saturnin'

def main():
    """Saturnin CLI manager.
    """
    parser: ArgumentParser = ArgumentParser(PROG_NAME, description=main.__doc__,
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    cmds = CommandManager(parser)
    cmds.load_commands('saturnin.commands.cli')
    cmds.run()

if __name__ == '__main__':
    main()




#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/cli.py
# DESCRIPTION:    Saturnin console commands
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin console commands


"""

from __future__ import annotations
from typing import Dict
from argparse import ArgumentParser, Namespace
from saturnin.base import Error
from saturnin.lib.command import Command, CommandManager

class StartCommand(Command):
    """START Saturnin console command.

    Starts Saturnin node or service.
    """
    def __init__(self):
        super().__init__('start', "Start Saturnin node or service")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        print(f"Running {self.name}")

class StopCommand(Command):
    """STOP Saturnin console command.

    Stops Saturnin node or service.
    """
    def __init__(self):
        super().__init__('stop', "Stop Saturnin node or service")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        print(f"Running {self.name}")

class ListCommand(Command):
    """LIST Saturnin console command.

    Lists runing nodes and services.
    """
    def __init__(self):
        super().__init__('list', "List runing nodes and services")
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        print(f"Running {self.name}")

